import hashlib
import zlib
from functools import partial
from pathlib import Path
from typing import IO, Callable, Optional, Union

from _hashlib import HASH as Hashobj

from .file import PathType, blockfileiter, iterfilelike, read_file

HashCls = Union[Callable[[], Hashobj], str]

FILE_IO_BUFFER_SIZE = 8 * 1024 * 1024


class HashobjCRC:  # don't inherit from Hashobj (TypeError: cannot create 'HashobjCRC' instances)
    digest_size = 4
    name = "crc"

    def __init__(self) -> None:
        self.prev = 0

    def update(self, data: bytes) -> None:
        """Update the hash object with the bytes-like object.
        Repeated calls are equivalent to a single call with the concatenation of all the arguments:
        m.update(a); m.update(b) is equivalent to m.update(a+b).
        """

        self.prev = zlib.crc32(data, self.prev)

    def digest(self) -> bytes:
        """Return the digest of the data passed to the update() method so far.
        This is a bytes object of size digest_size which may contain bytes in the whole range from 0 to 255.
        """

        return (self.prev & 0xFFFFFFFF).to_bytes(4, "big", signed=False)

    def hexdigest(self) -> str:
        """Like digest() except the digest is returned as a string object of double length,
        containing only hexadecimal digits. This may be used to exchange the value safely in email or other non-binary environments.
        """

        return self.digest().hex()

    def copy(self) -> "HashobjCRC":
        """Return a copy (“clone”) of the hash object.
        This can be used to efficiently compute the digests of data sharing a common initial substring.
        """

        out = HashobjCRC()
        out.prev = self.prev
        return out


def hash_file(
    path: PathType,
    hashcls: HashCls,
    chunk_size: int = FILE_IO_BUFFER_SIZE,
    mode: str = "rb",
    buffering: int = -1,
    encoding: Optional[str] = None,
    opener=None,
) -> Hashobj:
    # fixme: does this even work with `mode="rt"`?

    if isinstance(hashcls, str):
        m = hashlib.new(hashcls)
    else:
        m = hashcls()
    for d in blockfileiter(path, mode, buffering, encoding, chunk_size=chunk_size, opener=opener):
        m.update(d)
    return m


def hash_filelike(fr: IO[bytes], hashcls: HashCls, chunk_size: int = FILE_IO_BUFFER_SIZE) -> Hashobj:
    if isinstance(hashcls, str):
        m = hashlib.new(hashcls)
    else:
        m = hashcls()

    for d in iterfilelike(fr, chunk_size=chunk_size):
        m.update(d)
    return m


def hash_data(data: bytes, hashcls: HashCls) -> Hashobj:
    """Hashes `data` with `hashcls`.
    `hashcls` can either be a string like "md5" or a hash object like `hashlib.md5`.
    """

    if isinstance(hashcls, str):
        m = hashlib.new(hashcls)
    else:
        m = hashcls()
    m.update(data)
    return m


md4_hash_data = partial(hash_data, hashcls="md4")
md5_hash_data = partial(hash_data, hashcls=hashlib.md5)
sha1_hash_data = partial(hash_data, hashcls=hashlib.sha1)
crc_hash_data = partial(hash_data, hashcls=HashobjCRC)

md5_hash_file = partial(hash_file, hashcls=hashlib.md5)
sha1_hash_file = partial(hash_file, hashcls=hashlib.sha1)
crc32_hash_file = partial(hash_file, hashcls=HashobjCRC)  # output changed in v0.0.104


def hashsum_file_format(hashobj: Hashobj, path: str) -> str:
    return f"{hashobj.hexdigest()} *{path}"


ed2k_chunksize = 9728000


def ed2k_hash_file_v1(path: Path) -> str:
    """Returns ed2k hash.
    This hashing method is used by
    - MLDonkey
    - Shareaza
    - HashCalc

    This differs from `ed2k_hash_file_v2` only if the file size is a multiple of `ed2k_chunksize`.
    """

    if path.stat().st_size <= ed2k_chunksize:
        return md4_hash_data(read_file(path, "rb")).hexdigest()

    ed2k_hashes = (md4_hash_data(data).digest() for data in blockfileiter(path, "rb", chunk_size=ed2k_chunksize))

    return md4_hash_data(b"".join(ed2k_hashes)).hexdigest()


def ed2k_hash_file_v2(path: Path) -> str:
    """Returns ed2k hash.
    This hashing method is used by
    - eMule
    - AOM

    This differs from `ed2k_hash_file_v1` only if the file size is a multiple of `ed2k_chunksize`.
    """

    filesize = path.stat().st_size
    if filesize < ed2k_chunksize:
        return md4_hash_data(read_file(path, "rb")).hexdigest()

    ed2k_hashes = list(md4_hash_data(data).digest() for data in blockfileiter(path, "rb", chunk_size=ed2k_chunksize))

    if filesize % ed2k_chunksize == 0:
        ed2k_hashes.append(md4_hash_data(b"").digest())

    return md4_hash_data(b"".join(ed2k_hashes)).hexdigest()



from functools import reduce

import rhash

from .file import blockfileiter
from .hash import FILE_IO_BUFFER_SIZE
from .ops import bit_or


def multi_hash_file(path, hash_types, base=None, mode="rb", encoding=None, errors=None, chunk_size=FILE_IO_BUFFER_SIZE):
    """Hashes `path` with multiple hash functions.
    `hash_types` is a list of names, eg. ("CRC32", "SHA1").
    """

    hashdict = {
        "CRC32": rhash.CRC32,
        "MD4": rhash.MD4,
        "MD5": rhash.MD5,
        "SHA1": rhash.SHA1,
        "TIGER": rhash.TIGER,
        "TTH": rhash.TTH,
        "BTIH": rhash.BTIH,
        "ED2K": rhash.ED2K,
        "AICH": rhash.AICH,
        "WHIRLPOOL": rhash.WHIRLPOOL,
        "RIPEMD-160": rhash.RIPEMD160,
        "GOST12_256": rhash.GOST12_256,
        "GOST12_512": rhash.GOST12_512,
        "GOST94": rhash.GOST94,
        "GOST94_CRYPTOPRO": rhash.GOST94_CRYPTOPRO,
        "HAS-160": rhash.HAS160,
        "SNEFRU-128": rhash.SNEFRU128,
        "SNEFRU-256": rhash.SNEFRU256,
        "SHA-224": rhash.SHA224,
        "SHA-256": rhash.SHA256,
        "SHA-384": rhash.SHA384,
        "SHA-512": rhash.SHA512,
        "EDON-R256": rhash.EDONR256,
        "EDON-R512": rhash.EDONR512,
    }

    basedict = {
        "upper": rhash.RHPR_UPPERCASE,
        "raw": rhash.RHPR_RAW,
        "hex": rhash.RHPR_HEX,
        "HEX": rhash.RHPR_HEX | rhash.RHPR_UPPERCASE,
        "base32": rhash.RHPR_BASE32,
        "BASE32": rhash.RHPR_BASE32 | rhash.RHPR_UPPERCASE,
        "base64": rhash.RHPR_BASE64,
    }

    hashlist = tuple(hashdict[i.upper()] for i in hash_types)
    flags = basedict.get(base, 0)

    hasher = rhash.RHash(reduce(bit_or, hashlist))
    for data in blockfileiter(path, mode, encoding=encoding, errors=errors, chunk_size=chunk_size):
        hasher.update(data)
    hasher.finish()
    return tuple(hasher._print(i, flags) for i in hashlist)


if __name__ == "__main__":
    import argparse

    funcs = (
        "CRC32",
        "MD4",
        "MD5",
        "SHA1",
        "TIGER",
        "TTH",
        "BTIH",
        "ED2K",
        "AICH",
        "WHIRLPOOL",
        "RIPEMD-160",
        "GOST12_256",
        "GOST12_512",
        "GOST94",
        "GOST94_CRYPTOPRO",
        "HAS-160",
        "SNEFRU-128",
        "SNEFRU-256",
        "SHA-224",
        "SHA-256",
        "SHA-384",
        "SHA-512",
        "EDON-R256",
        "EDON-R512",
    )

    parser = argparse.ArgumentParser(
        description="Multi file hashing", epilog="Example: multihash C:\\Windows\\System32\\calc.exe -f CRC32 SHA1"
    )
    parser.add_argument(
        "-f",
        "--hashfuncs",
        nargs="+",
        help="list of hash functions. Can be one or more of: {}".format(", ".join(funcs)),
        required=True,
    )
    parser.add_argument("file", metavar="FILE", help="path to the file")
    args = parser.parse_args()
    print("\n".join(f"{i}: {j}" for i, j in zip(args.hashfuncs, multi_hash_file(args.file, args.hashfuncs))))



from typing import Any, Container, Hashable, Iterable, Iterator, Optional, Sequence, Tuple, Type, TypeVar, Union

from typing_extensions import Protocol  # typing.Protocol is available in Python 3.8+

ExceptionsType = Union[Type[Exception], Tuple[Type[Exception], ...]]
Number = Union[int, float]
T_co = TypeVar("T_co", covariant=True)


class HashableContainer(Protocol[T_co], Hashable, Container[T_co]):
    pass


class SizedIterable(Protocol[T_co]):
    def __len__(self) -> int: ...

    def __iter__(self) -> Iterator[T_co]: ...


class Comparable(Protocol):
    def __eq__(self, other: Any) -> bool: ...

    def __ne__(self, other: Any) -> bool: ...


class Orderable(Protocol):
    def __lt__(self, other: Any) -> bool: ...

    def __gt__(self, other: Any) -> bool: ...

    def __le__(self, other: Any) -> bool: ...

    def __ge__(self, other: Any) -> bool: ...


class Computable(Protocol):
    def __neg__(self) -> Any: ...

    def __add__(self, other: Any) -> Any: ...

    def __sub__(self, other: Any) -> Any: ...

    def __mul__(self, other: Any) -> Any: ...

    def __truediv__(self, other: Any) -> Any: ...

    def __floordiv__(self, other: Any) -> Any: ...

    def __abs__(self) -> Any: ...


class MutableComputable(Protocol):
    def __neg__(self) -> Any: ...

    def __add__(self, other: Any) -> Any: ...

    def __iadd__(self, other: Any) -> Any: ...

    def __sub__(self, other: Any) -> Any: ...

    def __isub__(self, other: Any) -> Any: ...

    def __mul__(self, other: Any) -> Any: ...

    def __imul__(self, other: Any) -> Any: ...

    def __truediv__(self, other: Any) -> Any: ...

    def __itruediv__(self, other: Any) -> Any: ...

    def __floordiv__(self, other: Any) -> Any: ...

    def __ifloordiv__(self, other: Any) -> Any: ...


class Cursor(Protocol):
    def callproc(self, procname: str, parameters: Optional[Sequence[Any]]) -> None: ...

    def close(self) -> None: ...

    def execute(self, operation: str, parameters: Optional[Sequence[Any]]) -> None: ...

    def executemany(self, operation: str, seq_of_parameters: Sequence[Sequence[Any]]) -> None: ...

    def fetchone(self) -> Optional[Sequence[Any]]: ...

    def fetchmany(self, size: Optional[int]) -> Sequence[Sequence[Any]]: ...

    def fetchall(self) -> Sequence[Sequence[Any]]: ...

    def nextset(self) -> Optional[bool]: ...

    def setinputsizes(self, sizes: Sequence[int]) -> None: ...

    def setoutputsize(self, size: int, column: Optional[int]) -> None: ...


class Connection(Protocol):
    def close(self) -> None: ...

    def commit(self) -> None: ...

    def rollback(self) -> None: ...

    def cursor(self) -> Cursor: ...


class CsvWriter(Protocol):
    def writerow(self, row: Iterable[Any]) -> Any: ...

    def writerows(self, rows: Iterable[Iterable[Any]]) -> None: ...



import re
from string import whitespace
from typing import Callable, Iterable, Iterator, Optional, Sequence, Union

from .iter import collapse_all, collapse_any
from .string import replace_multiple
from .unicode import unicode_categories
from .url import get_url_pattern, valid_uri_characters

uni_cats = unicode_categories()


def extract_urls(text: str) -> Iterator[str]:
    """Yields URLs from `text`."""

    pattern = r"https?\:\/\/[" + re.escape(valid_uri_characters) + r"]+"
    for match in re.finditer(pattern, text):
        url = match.group(0)
        yield url


def replace_typographical_punctuation(s: str) -> str:
    """Replaces typographical punctuation with ASCII punctuation."""

    d = {
        "\N{SINGLE LOW-9 QUOTATION MARK}": "'",
        "\N{DOUBLE LOW-9 QUOTATION MARK}": '"',
        "\N{LEFT SINGLE QUOTATION MARK}": "'",
        "\N{RIGHT SINGLE QUOTATION MARK}": "'",
        "\N{LEFT DOUBLE QUOTATION MARK}": '"',
        "\N{RIGHT DOUBLE QUOTATION MARK}": '"',
        "\N{FIGURE DASH}": "-",
        "\N{EN DASH}": "-",
        "\N{EM DASH}": "--",
        "\N{HORIZONTAL BAR}": "--",
        "\N{DOUBLE LOW LINE}": "_",
        "\N{HORIZONTAL ELLIPSIS}": "...",
    }
    return replace_multiple(s, d)


def newlines_to_spaces(s: str) -> str:
    """Replaces newline characters with spaces."""

    d = {
        "\r": "",
        "\n": " ",
        "\u0085": " ",  # should be \N{Next Line} or "\N{NEL}", but python 2.7 doesn't know that name
    }
    return replace_multiple(s, d)


def collapse_whitespace(s: str) -> str:
    """Collapses repeated whitespace into a single space character."""

    # separator_cats = ("Zl", "Zp", "Zs")
    # cats = tuple(uni_cats[i] for i in separator_cats)
    # separators = set.union(*cats)
    return "".join(collapse_all(s, set(whitespace), " "))


def collapse_space(s: str, space: str = " ") -> str:
    """Collapses repeated spaces into a single space character."""

    # slower: variant of this with join
    # slower: re.compile(r" {2,}").sub(" ", s)
    # slower: "".join(collapse_all(s, {" "}, " "))

    if s == space:
        return s

    ss = s.split(space)

    beg = space if not ss[0] else ""
    end = space if not ss[-1] else ""

    return beg + space.join(i for i in ss if i) + end


def collapse_punctuation_symbols(s: str) -> str:
    """Collapses any successive Unicode punctuation symbol in `s`."""

    punctuation_cats = ("Pc", "Pd", "Pe", "Pf", "Pi", "Po", "Ps")
    symbol_cats = ("Sc", "Sk", "Sm", "So")

    cats = tuple(uni_cats[i] for i in punctuation_cats + symbol_cats)
    unwantedset = set.union(*cats)
    return "".join(collapse_any(s, unwantedset))


class ReplaceURLs:
    def __init__(self, replacement: Union[str, Iterable], schemes: Optional[Sequence[str]] = None) -> None:
        if isinstance(replacement, str):
            self.repl: Union[str, Callable[[str], str]] = replacement
        else:
            self.repl = lambda match: next(replacement)  # type: ignore

        self.pattern = get_url_pattern(schemes)

    def __call__(self, s: str, count: int = 0) -> str:
        return self.pattern.sub(self.repl, s, count)



import heapq
import os
from collections.abc import Iterable
from contextlib import ExitStack
from itertools import islice
from operator import itemgetter
from typing import Any, AnyStr, Callable, Generic
from typing import Iterable as IterableT
from typing import Iterator, List, MutableSequence, Optional, Tuple, TypeVar, Union

from .file import copen
from .func import identity
from .math import argmax, argmin

T = TypeVar("T")
U = TypeVar("U")


def bubble_sort(seq: MutableSequence) -> None:
    """Slightly optimized version of stable bubble sort. Sorts input `seq` in place."""

    n = len(seq)
    while True:
        newn = 1
        for i in range(1, n):
            if seq[i - 1] > seq[i]:
                seq[i - 1], seq[i] = seq[i], seq[i - 1]
                newn = i
        if newn == 1:
            break
        n = newn


def selection_sort_min(seq: MutableSequence) -> None:
    # inplace, unstable

    n = len(seq)
    for i in range(n - 1):
        m = argmin(seq, i, n)
        if m != i:
            seq[i], seq[m] = seq[m], seq[i]


def selection_sort_max(seq: MutableSequence) -> None:
    # inplace, unstable

    for n in range(len(seq) - 1, 0, -1):
        m = argmax(seq[: n + 1])
        if m != n:
            seq[m], seq[n] = seq[n], seq[m]


def selection_sort_ll(seq):
    # inplace, stable, on linked lists
    raise NotImplementedError


class OptionalValue(Generic[T]):
    def __init__(self, value: T) -> None:
        self.value = value

    def __repr__(self) -> str:
        return "<Optional " + repr(self.value) + ">"

    def __eq__(self, other: "OptionalValue[T]") -> bool:
        return self.value == other.value

    def __lt__(self, other: "OptionalValue[T]") -> bool:
        a = self.value
        b = other.value

        if a is None:
            return True
        if b is None:
            return False

        if not isinstance(a, str) and isinstance(a, Iterable):
            a = tuple(map(OptionalValue, a))
        if not isinstance(b, str) and isinstance(b, Iterable):
            b = tuple(map(OptionalValue, b))

        return a < b


def sorted_index(it: IterableT[T], reverse: bool = False) -> Iterator[Tuple[T, int]]:
    """adds the position of each element after sorting without changing the order of the elements
    eg. [2, 6, -1, 4] -> [(2, 2), (6, 4), (-1, 1), (4, 3)]
    """

    return (
        (x[1][1], x[0])
        for x in sorted(enumerate(sorted(enumerate(it), key=lambda x: x[1], reverse=reverse), 1), key=lambda x: x[1][0])
    )


def sorted_by_list(
    tosort: IterableT[T], sortby: IterableT[U], reverse: bool = False, include_keys: bool = False
) -> Union[IterableT[Tuple[T, U]], Iterator[T]]:
    """Sorts `tosort` using the elements of `sortby`.
    The length of the shorter of both list determines the length of the output.
    eg. [1,2,3,4], [2,4,1,3] -> [3,1,4,2] for include_keys is False
    """

    ret = sorted(zip(tosort, sortby), key=itemgetter(1), reverse=reverse)
    if not include_keys:
        return map(itemgetter(0), ret)
    else:
        return ret


def sorted_with_keys(it: IterableT[T], key: Callable[[T], U], reverse: bool = False) -> Iterator[Tuple[T, U]]:
    """Sorts `it` using function `key` adds keys to output.
    eg. [2,4,1,3], lambda x: -x -> [(4,-4),(3,-3),(2,-2),(1,-1)]
    """

    with_keys = ((elm, key(elm)) for elm in it)
    return sorted(with_keys, key=itemgetter(1), reverse=reverse)


def external_sort(
    in_iterator: IterableT[AnyStr],
    temp_file_template: str,
    mode: str = "t",
    key: Optional[Callable[[str], Any]] = None,
    max_lines_temp: Optional[int] = None,
    max_lines_final: Optional[int] = None,
) -> Iterator[AnyStr]:
    """Sorts `in_iterator` by using external files. Keeps at most `max_lines_temp` lines in memory at any time.
    `in_iterator` must not yield any newlines. They must be escaped beforehand.
    If mode is "t" `in_iterator` must yield `str`, mode "b" it must yield `bytes`.
    `max_lines_final` limits output size. `key` is the key function for `sort()`.
    """

    # doesn't handle newlines in input correctly

    assert mode in {"t", "b"}

    if mode == "t":
        encoding: Optional[str] = "utf-8"
        errors: Optional[str] = "strict"
        newline: AnyStr = "\n"
    else:
        encoding = None
        errors = None
        newline = b"\n"

    if key is None:
        key = identity

    if max_lines_temp is None:
        # find some clever way to do that, based on memory and so on.
        max_lines_temp = 10000000

    def file_write_iter(path, it):
        with copen(path, "w" + mode, compresslevel=1, encoding=encoding, errors=errors, newline=newline) as fw:
            for line in it:
                fw.write(line + newline)

    # create temp files, write sequentially

    file_count = 0
    lines: List[AnyStr] = []

    for line in in_iterator:
        lines.append(line.rstrip(newline))  # removes newlines from input
        if len(lines) >= max_lines_temp:
            lines.sort(key=key)
            file_write_iter(temp_file_template.format(file_count), lines)
            file_count += 1
            lines = []
    if lines:
        lines.sort(key=key)
        file_write_iter(temp_file_template.format(file_count), lines)
        file_count += 1
        lines = []

    # merge temp files, read parallel

    filenames = [temp_file_template.format(i) for i in range(file_count)]
    with ExitStack() as stack:
        files = [
            stack.enter_context(copen(name, "r" + mode, encoding=encoding, errors=errors, newline=newline))
            for name in filenames
        ]

        merged = heapq.merge(*files, key=key)

        for line in islice(merged, max_lines_final):
            yield line.rstrip(newline)

    # remove temp files

    for name in filenames:
        os.remove(name)


def external_sort_file(
    in_file: str,
    out_file: str,
    temp_file_template: str,
    sort_key: Optional[Callable[[str], Any]] = None,
    max_lines_temp: Optional[int] = None,
    max_lines_final: Optional[int] = None,
) -> None:
    """Sorts file `in_file` to `out_file` line by line using external files
    provided by filename `template temp_file_template`.
    At most `max_lines_temp` lines will be in memory at any time.
    """

    with copen(in_file, "rt", encoding="utf-8", errors="strict") as fr, copen(
        out_file, "wt", encoding="utf-8", errors="strict", newline="\n"
    ) as fw:
        lines = external_sort(fr, temp_file_template, "t", sort_key, max_lines_temp, max_lines_final)

        fw.writelines(map(lambda s: s + "\n", lines))



import json
from itertools import chain

from setuptools import setup

# nltk.download("punkt")

with open("README.md", encoding="utf-8") as fr:
    long_description = fr.read()

with open("install_requires.json", encoding="utf-8") as fr:
    install_requires = json.load(fr)

with open("extras_require.json", encoding="utf-8") as fr:
    extras_require = json.load(fr)

extras_require["all"] = sorted(set(chain.from_iterable(extras_require.values())))

setup(
    author="Dobatymo",
    name="genutility",
    version="0.0.107",
    url="https://github.com/Dobatymo/genutility",
    description="A collection of various Python utilities",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: ISC License (ISCL)",
        "Operating System :: OS Independent",
        "Topic :: Utilities",
    ],
    packages=["genutility", "genutility.compat", "genutility.fileformats", "genutility.win"],
    package_data={"genutility": ["py.typed", "data/*.tsv"]},
    python_requires=">=3.7",
    install_requires=install_requires,
    extras_require=extras_require,
    license_files=["LICENSE"],
)



from math import exp, sqrt
from typing import Any, Callable, Dict, Generic, Iterator, Optional, Sequence, Tuple, TypeVar

import numpy as np

# from .numba import opjit

T = TypeVar("T")

RGB_YELLOW = (255, 255, 0)
RGB_WHITE = (255, 255, 255)


def assert_square(name: str, value: np.ndarray) -> None:
    """Raises a ValueError if matrix `value` is not square.
    value: [A, A]
    """

    if not len(value.shape) == 2 or value.shape[0] != value.shape[1]:
        raise ValueError(f"{name} must be a square")


def normalized_choice(
    p_ind, p_prob
):  # is this neccessary? I think `np.random.choice` can handle unnormalized probabilities
    return np.random.choice(p_ind, p=p_prob / np.sum(p_prob))


def shannon_entropy(ps: np.ndarray, base: int = 2) -> float:
    """Calculates the Shannon entropy for probabilities `ps` with `base`."""

    return -np.sum(ps * np.log(ps) / np.log(base))


def is_rgb(img: np.ndarray) -> bool:
    """Simply tests if `img` has 3 channels."""

    return len(img.shape) >= 1 and img.shape[-1] == 3


# @opjit() # np.errstate doesn't work in numba...
def rgb_to_hsi(image: np.ndarray) -> np.ndarray:
    """Converts an array [..., channels] of RGB values to HSI color values (H in rad).
    RGB values are assumed to be normalized to (0, 1).
    """

    if not is_rgb(image):
        raise ValueError("Input needs to be an array of RGB values")

    r = image[..., 0]
    g = image[..., 1]
    b = image[..., 2]

    out = np.zeros_like(image)

    # allequal = (img == img[:, :, 0, np.newaxis]).all(axis=-1)

    with np.errstate(invalid="ignore"):
        tmp = (2.0 * r - g - b) / 2.0 / np.sqrt((r - g) ** 2 + (r - b) * (g - b))  # if r==g==b then 0/0

        theta = np.arccos(np.clip(tmp, -1.0, +1.0))

        out[..., 0] = np.where(b <= g, theta, 2 * np.pi - theta)  # H
        out[..., 2] = np.sum(image, axis=-1) / 3.0  # I
        out[..., 1] = 1 - np.amin(image, axis=-1) / out[..., 2]  # S if r==g==b==0 then 0/0

    np.nan_to_num(out[..., 0:2], copy=False)

    return out


# @opjit() # np.dot with more than 2 dimensions is not supported...
def rgb_to_ycbcr(image: np.ndarray) -> np.ndarray:
    """Converts an array [..., channels] of RGB values to Digital Y'CbCr (0-255).
    RGB values are assumed to be normalized to (0, 1).
    Don't forget to cast to uint8 for pillow.
    """

    """  from RGB (0-1).
    """

    if not is_rgb(image):
        raise ValueError("Input needs to be an array of RGB values")

    m = np.array(
        [
            [+065.481, +128.553, +024.966],
            [-037.797, -074.203, +112.000],
            [+112.000, -093.786, -018.214],
        ]
    )
    a = np.array([16, 128, 128])

    return np.dot(image, m.T) + a


def random_triangular_matrix(size: int, lower: bool = True) -> np.ndarray:
    """Returns a triangular matrix with random value between 0 and 1 uniformly."""

    a = np.random.uniform(0, 1, (size, size))
    if lower:
        ind = np.triu_indices(5, 1)
    else:
        ind = np.tril_indices(5, 1)
    a[ind] = 0

    return a


def is_square(A: np.ndarray) -> bool:
    if len(A.shape) != 2:
        return False

    if A.shape[0] != A.shape[1]:
        return False

    return True


def batch_vTAv(A: np.ndarray, v: np.ndarray) -> np.ndarray:
    """Performs batched calculation of `v^T A v` transform.
    Special case of bilinear form `x^T A y`

    A: float[B, X, X]
    v: float[B, X]

    returns: float[B]
    """

    """ Faster than
        Av = np.matmul(A, v[...,:,None]) # [B, X, 1]
        return np.matmul(v[...,None,:], Av).squeeze((-2, -1)) # [B]
    """

    return np.einsum("...k,...kl,...l->...", v, A, v)


def batch_inner(a: np.ndarray, b: np.ndarray, verify: bool = True) -> np.ndarray:
    """Performs a batched inner product over the last dimension.
    Replacement for deprecated `from numpy.core.umath_tests import inner1d`.
    Shapes: (B, X), (B, X) -> (B, )
    """

    if verify and a.shape != b.shape:
        raise ValueError("All dimensions have to be equal")

    if a.shape[-1] == 0:
        return np.empty_like(a)

    return np.einsum("...i,...i->...", a, b)  # faster than np.sum(a * b, axis=-1)


def batch_outer(a: np.ndarray, b: np.ndarray, verify: bool = True) -> np.ndarray:
    """Performs a batched outer product over the last dimension.
    Shapes: (B, X), (B, Y) -> (B, X, Y)
    """

    if verify and a.shape[:-1] != b.shape[:-1]:
        raise ValueError("All except the last dimension have to be equal")

    return np.einsum("...i,...j->...ij", a, b)  # slightly faster than np.multiply(a[...,:,None], b[...,None,:])


def batchtopk(
    probs: np.ndarray, k: Optional[int] = None, axis: int = -1, reverse: bool = False
) -> Tuple[np.ndarray, np.ndarray]:
    """`probs` values ndarray
    `k` take the smallest `k` elements, if `reverse` is False
            and the largest `k` if `reverse` is True
    `axis` sorting and selection axis.
    """

    if k is not None and k <= 0:
        raise ValueError("k must be larger than 0. Use None to chose all elements.")

    if axis != -1:
        raise ValueError("Only last axis supported atm")

    if len(probs.shape) <= 1:
        raise ValueError("probs must be at least 2-dimensional")

    if reverse:
        sign = -1
    else:
        sign = 1

    indices = np.argsort(sign * probs, axis=-1)  # use argpartition?
    probs = np.take_along_axis(probs, indices[..., :k], axis=-1)

    return indices, probs


# @opjit() # np.diagonal not supported by numba as of version 0.52.0
def logtrace(m: np.ndarray) -> np.ndarray:
    """Calcuates the sum of the logs of the diagonal elements (batchwise if necessary)
    m: [..., x, x]
    """

    """ note: performance cannot easily be improve by numba.
        `np.diagonal` not supported by numba 0.52.0
    """

    return np.sum(np.log(np.diagonal(m, axis1=-2, axis2=-1)), axis=-1)


def shiftedexp(pvals: np.ndarray) -> np.ndarray:
    """Shifts `pvals` by the largest value in the last dimension
    before the exp is calculated to prevent overflow (batchwise if necessary).
    Can be used if probabilities are normalized again later.
    """

    if pvals.shape[-1] == 0:
        return np.empty_like(pvals)

    return np.exp(pvals - np.amax(pvals, axis=-1)[..., None])


class Sampler:
    """Sample from discrete CDF."""

    def __init__(self, cdf: np.ndarray) -> None:
        self.cdf = cdf
        self.psum = cdf[-1]

    def __call__(self) -> int:
        """Sample one."""

        rand = np.random.uniform(0, self.psum)
        return np.searchsorted(self.cdf, rand, side="right")

    def sample(self, n: int) -> np.ndarray:
        """Sample `n`."""

        rands = np.random.uniform(0, self.psum, n)
        return np.searchsorted(self.cdf, rands, side="right")

    def pdf(self, n: int, minlength: Optional[int] = None) -> np.ndarray:
        out = np.bincount(self.sample(n), minlength=minlength)
        return out / n


def sample_probabilities(pvals: np.ndarray) -> Callable[[], int]:
    """Sample from list of probabilities `pvals` with replacement.
    The probabilities don't need to be normalized.
    """

    return Sampler(np.cumsum(pvals))


class UnboundedSparseMatrix(Generic[T]):
    def __init__(self, dtype: type = float) -> None:
        self.dtype = dtype
        self.zero = self.dtype(0)
        self.m: Dict[Tuple[int, int], T] = {}
        self.cols = 0
        self.rows = 0

    def __getitem__(self, slice: Tuple[int, int]) -> T:
        return self.m.get(slice, self.zero)

    def __setitem__(self, slice: Tuple[int, int], value: T) -> None:
        c, r = slice
        self.cols = max(self.cols, c + 1)
        self.rows = max(self.rows, r + 1)
        self.m[slice] = value

    def todense(self) -> np.ndarray:
        ret = np.zeros((self.cols, self.rows), dtype=self.dtype)

        for slice, value in self.m.items():
            ret[slice] = value

        return ret


def normalize(pvals: np.ndarray) -> np.ndarray:
    return pvals / np.sum(pvals)


def categorical(pvals: np.ndarray) -> int:
    """Sample from the categorical distribution using `pvals`.
    See: https://en.wikipedia.org/wiki/Categorical_distribution
    """

    return sample_probabilities(pvals)()  # faster than: np.argmax(np.random.multinomial(1, normalize(pvals)))


def population2cdf(population: np.ndarray) -> np.ndarray:
    """Convert a population (list of observations) to a CDF."""

    population = np.sort(population)
    return np.searchsorted(population, population, side="right") / len(population)


def pmf2cdf(pdf: np.ndarray) -> np.ndarray:
    """Convert a discrete PDF into a discrete CDF."""

    cdf = np.cumsum(pdf)
    return cdf / cdf[-1]


def _two_sample_kolmogorov_smirnov_same_length(
    cdf1: np.ndarray, cdf2: np.ndarray, n1: int, n2: int
) -> Tuple[float, float]:
    # note: yields different results as `scipy.stats.ks_2samp`

    if len(cdf1) != len(cdf2):
        raise ValueError("Both CDFs must have the same length")

    D = np.amax(np.abs(cdf1 - cdf2))  # K-S statistic
    level = exp(-2 * (D / sqrt((n1 + n2) / (n1 * n2))) ** 2)
    return D, level


def _two_sample_kolmogorov_smirnov_population(
    p1: np.ndarray, p2: np.ndarray, alpha: float = 0.05
) -> Tuple[float, float, bool]:
    # note: yields different results as `scipy.stats.ks_2samp`

    cdf1 = population2cdf(p1)
    cdf2 = population2cdf(p2)

    statistic, pvalue = _two_sample_kolmogorov_smirnov_same_length(cdf1, cdf2, len(cdf1), len(cdf2))
    reject = pvalue < alpha
    return statistic, pvalue, reject


def _two_sample_kolmogorov_smirnov_pmf(
    pmf1: np.ndarray, pmf2: np.ndarray, alpha: float = 0.05
) -> Tuple[float, float, bool]:
    # note: yields different results as `scipy.stats.ks_2samp`
    """Tests the null hypothesis that both samples belong to the same distribution.
    See: https://en.wikipedia.org/wiki/Kolmogorov%E2%80%93Smirnov_test#Two-sample_Kolmogorov%E2%80%93Smirnov_test
    """

    cdf1 = np.cumsum(pmf1)
    cdf2 = np.cumsum(pmf2)

    n1 = cdf1[-1]
    n2 = cdf2[-1]

    # cannot be inplace because of type conversion
    cdf1 = cdf1 / n1
    cdf2 = cdf2 / n2

    statistic, pvalue = _two_sample_kolmogorov_smirnov_same_length(cdf1, cdf2, n1, n2)
    reject = pvalue < alpha
    return statistic, pvalue, reject


def inf_matrix_power(pm: np.ndarray, dtype=np.float64) -> np.ndarray:
    """Calculate stochastic matrix `pm` to the power of infinity,
    by finding the eigenvector which corresponds to the eigenvalue 1.
    """

    w, v = np.linalg.eig(
        pm
    )  # scipy.linalg.eig would probably by faster as it can return the left and right eigen vectors

    if not np.isclose(w[0], 1.0):
        raise ValueError("The first eigenvalue is not none. Is this a right stochastic matrix?")

    vi = np.linalg.inv(v)
    d = np.zeros(pm.shape[0], dtype=dtype)
    d[0] = 1.0

    return np.matmul(v, np.matmul(np.diag(d), vi))


def decompress(selectors: np.ndarray, data: np.ndarray, default) -> np.ndarray:
    """Same result as:
    `np.array(list(genutility.iter.decompress(selectors, iter(data), default)))`
    but faster.
    `selectors` should be a bool array.
    """

    out = np.full(len(selectors), default)
    out[selectors] = data
    return out


def unblock(arr: np.ndarray, n1: int, n2: int, axis1: int = -1, axis2: int = -2, blocksize: bool = False) -> np.ndarray:
    """Inverse of np.block.
    Set axis to (-2, -1) to modify the order of the result.
    """

    """ test (stackoverflow): Ok, so considering I have N block matrices with bm x bn dimension and want to stack them in a m x n matrix, provided N = m x n, I would then have x.reshape(m,n,bm,bn).swapaxes(1,2).reshape(bm*m,-1)
    """

    s = np.array(arr.shape)
    if s[axis1] % n1 != 0 or s[axis2] % n2 != 0:
        raise ValueError(f"{s[axis1]}x{s[axis2]} does not divide by {n1}x{n2}")

    if blocksize:
        n1 = s[axis1] // n1
        n2 = s[axis2] // n2

    # this first .split adds a new dimensions on the outside, so if a absolute index
    # is given for the second axis it must be moved one to the right
    if axis2 >= 0:
        _axis2 = axis2 + 1
    else:
        _axis2 = axis2

    arr = np.array(np.split(arr, n1, axis1))
    arr = np.array(np.split(arr, n2, _axis2))

    inv_blocksize = n1 * n2
    total = s[axis1] * s[axis2]
    s[axis2] = inv_blocksize
    s[axis1] = total // inv_blocksize

    return np.reshape(arr, s)


def unblock2d(arr: np.ndarray, n1: int, n2: int, blocksize: bool = False) -> np.ndarray:
    """
    arr: [width, height, ...]

    returns: (block_num, block_size)
    """

    s = np.array(arr.shape)
    if blocksize:
        bs1 = n1
        bs2 = n2
        bn1 = s[0] // n1
        bn2 = s[1] // n2
    else:
        bn1 = n1
        bn2 = n2
        bs1 = s[0] // n1
        bs2 = s[1] // n2

    if len(arr.shape) > 2:
        return arr.reshape(bn1, bs1, bn2, bs2, -1).swapaxes(1, 2).reshape(bn1 * bn2, bs1 * bs2, -1)
    else:
        return arr.reshape(bn1, bs1, bn2, bs2).swapaxes(1, 2).reshape(bn1 * bn2, bs1 * bs2)


def block2d(arr: np.ndarray, bn1: int, bn2: int, bs1: int, bs2: int) -> np.ndarray:
    """
    arr: (block_num, block_size)

    returns: (width, height)
    """

    s = np.array(arr.shape)
    assert s[0] == bn1 * bn2
    assert s[1] == bs1 * bs2

    return arr.reshape(bn1, bn2, bs1, bs2).swapaxes(1, 2).reshape(bn1 * bs1, bn2 * bs2)


def remove_color(img: np.ndarray, ratio: float, neutral_color: Tuple[int, int, int] = RGB_WHITE) -> None:
    """Replace colored pixels with a `neutral_color`. The `ratio` defines the 'colorfulness' above
    which level the pixel should be replace.
    I.e. if the `ratio` is 1 nothing will be replaced,
    if `ratio` is 0 only strict greys are kept unmodified.
    """

    channels = img.shape[-1]
    assert channels == 3, "Not a 3 channel color image"

    norm = np.std(np.array(RGB_YELLOW))  # this is the same for all pure colors

    sd = np.std(img, axis=-1)
    img[sd > ratio * norm] = neutral_color


# @opjit(cache=False) doesn't help a lot, because it's a generator
def sliding_window_2d(
    image: np.ndarray, window_size: Tuple[int, int], step_size: Tuple[int, int] = (1, 1)
) -> Iterator[np.ndarray]:
    win_x, win_y = window_size
    step_x, step_y = step_size
    height, width = image.shape[0:2]

    for y in range(0, height - win_y + 1, step_y):
        for x in range(0, width - win_x + 1, step_x):
            yield image[y : y + win_y, x : x + win_x, ...]


def histogram_correlation(hist1: np.ndarray, hist2: np.ndarray) -> np.ndarray:
    """Input shape of `hist1` and `hist2`: [batch..., levels]. The correlation is calculated over `levels`
    and then batched over the remaining dimensions.
    Similar to `cv2.compareHist(hist1, hist2, cv2.HISTCMP_CORREL)`.
    """

    if hist1.ndim < 1 or hist1.shape[-1] < 1 or hist1.shape != hist2.shape:
        raise ValueError(f"Histogram shapes unsupported: {hist1.shape}, {hist2.shape}")

    h1norm = (hist1.T - np.mean(hist1, axis=-1)).T
    h2norm = (hist2.T - np.mean(hist2, axis=-1)).T

    if np.count_nonzero(h1norm) == 0 or np.count_nonzero(h2norm) == 0:
        raise ValueError("Both histograms are completely flat")

    num = np.sum(h1norm * h2norm, axis=-1)
    denom = np.sqrt(np.sum(h1norm**2, axis=-1) * np.sum(h2norm**2, axis=-1))
    return num / denom


def bincount_batch(x: np.ndarray, axis: int = -1, minlength: int = 0) -> np.ndarray:
    if x.shape[axis] == 0:
        raise ValueError("Specified axis of x cannot be 0")

    minlength = max(minlength, x.max() + 1)
    return np.apply_along_axis(np.bincount, axis, x, minlength=minlength)


def stochastic(x: np.ndarray) -> np.ndarray:
    """It normalizes the last dimension of an ndarray to sum to 1.
    It can be used to convert (batches of) vectors to stochastic vectors
    or (batches of) matrices to *right* stochastic matrices.
    Right stochastic matrices are also called transitions matrices.
    """

    n = np.linalg.norm(x, 1, axis=-1, keepdims=True)
    # n = np.sum(x, axis=-1, keepdims=True) # todo: same result (except dtype), which is faster?

    with np.errstate(invalid="raise"):  # see: `normalized`
        return x / n


def sequence_mask(lengths: np.ndarray, maxlen: Optional[int] = None, dtype: Any = None) -> np.ndarray:
    """cf. tf.sequence_mask
    lengths: [N]
    """

    if not lengths:
        return np.array([], dtype=dtype or np.bool_)

    if maxlen is None:
        maxlen = max(lengths)
    row_vector = np.arange(maxlen)
    matrix = np.expand_dims(lengths, -1)

    ret = row_vector < matrix
    if dtype is not None:
        ret = ret.astype(dtype)

    return ret


# @opjit() # doesn't work atm because numba doesn't support `None` indexing
def _viterbi_dense_masked(
    p_emit: np.ndarray, p_trans: np.ndarray, p_trans0: np.ndarray, mask: np.ndarray
) -> np.ndarray:
    batch_size, T, N = p_emit.shape

    N1, N2 = p_trans.shape
    assert N == N1 == N2

    assert (batch_size, T) == mask.shape
    assert (N,) == p_trans0.shape

    trellis = np.zeros((T, batch_size, N), dtype=p_emit.dtype)
    states = np.zeros((T, batch_size, N), dtype=np.intp)

    # even if sequences of length 0 should be allowed, this does not have to be masked, as there cannot be a wrong result
    trellis[0] = p_trans0 + p_emit[:, 0, :]  # broadcast p_trans0: [N] -> batch_size*[[N]]

    for t in range(1, T):
        masked_p_trans = (
            mask[:, t, None, None] * p_trans[None, :, :]
        )  # [batch_size, 1, 1] * [1, N, N] == [batch_size, N, N]
        weighted_scores = (
            trellis[t - 1, :, :, None] + masked_p_trans
        )  # [batch_size, N, N] # scores and p_trans broadcasted
        max_scores = np.amax(weighted_scores, axis=1)  # [batch_size, N]
        trellis[t] = max_scores + p_emit[:, t, :]  # [batch_size, N] remember highest score of each path
        states[t] = np.argmax(
            weighted_scores, axis=1
        )  # [batch_size, N] remember index of best path, should be repeated for padding vals

    tokens = np.zeros((T, batch_size), dtype=np.intp)
    tokens[T - 1] = np.argmax(trellis[T - 1], axis=1)  # [batch_size]

    for t in range(T - 1, 0, -1):
        tokens[t - 1] = states[t, np.arange(batch_size), tokens[t]]  # [batch_size]

    return tokens.T  # [batch_size, T]


def viterbi_dense(
    p_emit: np.ndarray, p_trans: np.ndarray, p_trans0: Optional[np.ndarray] = None, mask: Optional[np.ndarray] = None
) -> np.ndarray:
    """Viterbi algorithm for finding the optimal path.
    One square transition matrix can be specified.

    T: max sequence length
    N: vocab size
    batch_size: to vectorize

    p_emit: masked emission scores [batch_size, T, N] float
    p_trans: transition scores [N, N] float
    p_trans0: initial transition scores [N] float
    mask: mask out padding values in transitions [batch_size, T] float
    """

    batch_size, T, N = p_emit.shape

    if mask is None:
        mask = np.ones((batch_size, T), dtype=p_trans.dtype)

    if p_trans0 is None:
        p_trans0 = np.zeros(N, dtype=p_emit.dtype)

    return _viterbi_dense_masked(p_emit, p_trans, p_trans0, mask)


# @opjit() # doesn't work atm because numba cannot store arbitrary elements in lists I think
def viterbi_sparse(p_emit: Sequence[np.ndarray], p_trans: Sequence[np.ndarray]) -> np.ndarray:
    """Viterbi algorithm for finding the optimal path.
    The number of emission probabilities per index can vary
    and a separate matrix can be specified for each transition.

    p_emit: Sequence of [x] matrices, where x can vary for every entry
    p_trans: Sequence of [y, z] matrices, where y and z can vary for every entry

    todo: using `Iterable`s instead of Sequence`s should be possible.
    """

    T = len(p_emit)

    assert T - 1 == len(p_trans)

    trellis = [p_emit[0]]
    states = [None]

    for t in range(1, T):
        weighted_scores = trellis[-1][:, None] + p_trans[t - 1]  # [x, y] # scores and p_trans broadcasted
        max_scores = np.amax(weighted_scores, axis=0)  # [y]
        trellis.append(np.add(max_scores, p_emit[t]))  # [y] remember highest score of each path
        states.append(np.argmax(weighted_scores, axis=0))  # [y] remember index of best path

    assert len(trellis) == T and len(states) == T

    tokens = [None] * T  # [T]
    tokens[-1] = np.argmax(trellis[-1], axis=0)  # []

    for t in range(T - 1, 0, -1):
        tokens[t - 1] = states[t][tokens[t]]  # []

    return tokens


_bit_counts = np.array([int(bin(x).count("1")) for x in range(256)]).astype(np.uint8)


def hamming_dist_packed(a: np.ndarray, b: np.ndarray, axis: Optional[int] = -1) -> np.ndarray:
    return np.sum(_bit_counts[np.bitwise_xor(a, b)], axis=axis)


def get_num_chunks(shape: np.ndarray, chunksize: np.ndarray) -> int:
    return np.prod(np.ceil(shape / chunksize).astype(np.int_))


def broadcast_shapes(*shapes: Tuple[int, ...]) -> Tuple[int, ...]:
    """np.broadcast_shapes requires `numpy==1.20.0`,
    which is not available for `python < 3.7`.
    """

    arrays = [np.empty(shape) for shape in shapes]
    return np.broadcast(*arrays).shape


def hamming_dist_packed_chunked(
    a: np.ndarray, b: np.ndarray, chunksize: Tuple[int, ...], axis: Optional[int] = -1
) -> Iterator[Tuple[Tuple[int, ...], np.ndarray]]:
    outshape = broadcast_shapes(a.shape, b.shape)
    select_broadcasted_axis = slice(0, 1)

    if len(outshape) - len(chunksize) != 1:
        raise ValueError("Length of `chunksize` must be one less the number of input dimensions")

    if len(outshape) == 2:
        for x in range(0, outshape[0], chunksize[0]):
            if a.shape[0] != outshape[0]:
                aix = select_broadcasted_axis
            else:
                aix = slice(x, x + chunksize[0])

            if b.shape[0] != outshape[0]:
                bix = select_broadcasted_axis
            else:
                bix = slice(x, x + chunksize[0])

            yield (x,), hamming_dist_packed(a[aix, :], b[bix, :], axis=axis)

    elif len(outshape) == 3:
        for x in range(0, outshape[0], chunksize[0]):
            for y in range(0, outshape[1], chunksize[1]):
                if a.shape[0] != outshape[0]:
                    aix = select_broadcasted_axis
                else:
                    aix = slice(x, x + chunksize[0])

                if b.shape[0] != outshape[0]:
                    bix = select_broadcasted_axis
                else:
                    bix = slice(x, x + chunksize[0])

                if a.shape[1] != outshape[1]:
                    aiy = select_broadcasted_axis
                else:
                    aiy = slice(y, y + chunksize[1])

                if b.shape[1] != outshape[1]:
                    biy = select_broadcasted_axis
                else:
                    biy = slice(y, y + chunksize[1])

                yield (x, y), hamming_dist_packed(a[aix, aiy, :], b[bix, biy, :], axis=axis)

    else:
        raise ValueError("Input must either be 2 or 3 dimensional")


def hamming_dist(a: np.ndarray, b: np.ndarray, axis: Optional[int] = -1) -> np.ndarray:
    return np.count_nonzero(a != b, axis=axis)


def center_of_mass_2d(arr: np.ndarray, dtype=np.float32) -> np.ndarray:
    """Batched center of mass calculation of 2d arrays
    `arr`: [..., x, y]
    """

    total = np.sum(arr, axis=(-1, -2))
    grids = np.ogrid[[slice(0, i) for i in arr.shape[-2:]]]
    with np.errstate(invalid="ignore"):
        results = np.array([np.sum(arr * grid.astype(dtype), axis=(-1, -2)) / total for grid in grids], dtype=dtype)
    results = np.moveaxis(results, 0, -1)
    return results


def ceildiv(a: np.ndarray, b: np.ndarray) -> np.ndarray:
    return -(a // -b)


def image_grid(images: np.ndarray, nrow: int = 8, fill_value=(0, 0, 0)) -> np.ndarray:
    """Creates a picture grid (left to right, top to bottom).
    `images`: Input array of shape [n, height, widght, channel]
    `nrow`: Number of images per row
    `fill_value`: Pad empty space with pixels of shape [channel]
    """

    if not images.ndim == 4:
        raise ValueError("Input shape must be [n, height, width, channel]")

    h, w = images.shape[1:3]
    nbatch = images.shape[0]
    channels = images.shape[3]
    rows = ceildiv(nbatch, nrow)
    cols = min(nbatch, nrow)

    if channels != len(fill_value):
        raise ValueError("fill_value must match number of channels")

    out = np.full((h * rows, w * cols, channels), fill_value, dtype=images.dtype)

    for i in range(images.shape[0]):
        row = i // cols
        col = i % cols
        out[h * row : h * (row + 1), w * col : w * (col + 1), :] = images[i]

    return out



import re
import socket
from typing import Optional

import netifaces


def is_ipv4(s: str) -> bool:
    """Tests if `s` is a IPv4 address string.
    It only validates the common `x.x.x.x` format and rejects less common ones like `x.x`.
    """

    try:
        nums = tuple(map(int, s.split(".")))
        return len(nums) == 4 and min(nums) >= 0 and max(nums) <= 255
    except (AttributeError, ValueError):
        return False


simple_email_regex = r"[^@\s]+@[^@\s]+\.[^@\s]+"


def is_email(s: str) -> bool:
    """Coarse check to test of a string is an email. Will accept some invalid ones like
    "asd@.asd.com" and reject some valid ones like "asd@localhost"
    """

    return re.fullmatch(simple_email_regex, s) is not None


def get_standard_gateway(default: Optional[str] = None) -> Optional[str]:
    """Returns the standard IPv4 gateway."""

    for ipv4_gateway in netifaces.gateways()["default"][netifaces.AF_INET]:
        try:
            socket.inet_aton(ipv4_gateway)
            return ipv4_gateway
        except OSError:
            pass

    return default



from typing import Any, Callable, Dict, Iterable, Mapping, Optional

from werkzeug.exceptions import BadRequest
from werkzeug.wsgi import get_host

WsgiApp = Any


class HostDispatcher:
    """Dispatch WSGI apps by host.
    app = HostDispatcher({'a.com': app_a, 'b.com': app_b})
    """

    def __init__(self, hosts: Dict[str, WsgiApp], default: Optional[WsgiApp] = None) -> None:
        self.hosts = hosts
        self.default = default or BadRequest()  # rfc7230#section-5.4

    def __call__(self, environ: Mapping[str, Any], start_response: Callable) -> Iterable[bytes]:
        host = get_host(environ)
        app = self.hosts.get(host, self.default)
        return app(environ, start_response)



from typing import Callable, Iterable, List, Optional

import numpy as np

Tokenizer = Callable[[str], Iterable[str]]


def hamming_distance(a: bytes, b: bytes) -> int:
    a = np.unpackbits(np.frombuffer(a, dtype=np.uint8))
    b = np.unpackbits(np.frombuffer(b, dtype=np.uint8))
    return np.count_nonzero(a != b)


def default_tokenizer(text: str) -> List[str]:
    return text.lower().split()


def same_words_similarity(a: str, b: str, tokenizer: Optional[Tokenizer] = None) -> int:
    tokenizer = tokenizer or default_tokenizer

    set_a = set(tokenizer(a))
    set_b = set(tokenizer(b))

    return len(set_a & set_b)



from functools import reduce
from operator import mul
from typing import Iterable, TypeVar

from ..typing import Computable

T = TypeVar("T")


try:
    from math import prod  # New in version 3.8

except ImportError:

    def prod(it: Iterable[T], start: T = 1) -> T:  # type: ignore
        """Counterpart to built-in `sum()`."""

        return reduce(mul, it, start)

    def prod_2(it: Iterable[Computable], start: Computable = 1) -> Computable:
        """Counterpart to built-in `sum()`."""

        for x in it:
            start *= x
        return start


try:
    from math import dist  # New in version 3.8
except ImportError:
    from math import sqrt

    def dist(p, q):
        return sqrt(sum((px - qx) ** 2.0 for px, qx in zip(p, q)))


try:
    from math import isqrt  # New in version 3.8

except ImportError:
    try:
        import gmpy2

        def isqrt(n):
            """Integer square root of `n`."""

            return int(gmpy2.isqrt(gmpy2.mpz(n)))

    except ImportError:

        def isqrt(n):
            """Integer square root of `n`.
            It uses float conversion and is not accurate for large numbers.
            Install `gmpy2` for improved speed and accuracy.
            """

            return int(n**0.5)



import os.path
from configparser import ConfigParser
from importlib.util import find_spec
from typing import Any, Dict, Optional, Union

from .os import get_appdata_dir


def sort_config(inpath: str, outpath: str) -> None:
    from sortedcontainers import SortedDict

    config = ConfigParser(dict_type=SortedDict)
    config.read(inpath, encoding="utf-8")
    with open(outpath, "w", encoding="utf-8") as fw:
        config.write(fw)


def _load(*names: str) -> Dict[str, Any]:
    from .toml import read_toml

    name = names[-1]
    configfilename = name + ".toml"

    # try appdata directory
    try:
        appdata_path = os.path.join(get_appdata_dir(), *names, configfilename)
        return read_toml(appdata_path)
    except FileNotFoundError:
        pass

    # try module directory
    try:
        spec = find_spec(name)

        if spec is None:
            raise ImportError(f"No module named '{name}'")  # or FileNotFoundError?

        if spec.has_location:
            assert spec.origin  # for mypy
            modpath = os.path.dirname(spec.origin)
        else:
            try:
                modpath = spec.submodule_search_locations[0]  # type: ignore
            except (TypeError, IndexError):
                raise FileNotFoundError

        return read_toml(os.path.join(modpath, configfilename))
    except (ImportError, FileNotFoundError):
        pass

    # try working directory
    try:
        return read_toml(configfilename)
    except FileNotFoundError:
        raise FileNotFoundError(
            f"{configfilename} could not be found in application data, module path or current directory"
        )


def load(*names: str, json_schema: Optional[Union[dict, str]] = None) -> Dict[str, Any]:
    from .json import read_json_schema

    obj = _load(*names)

    if json_schema is None:
        return obj

    from jsonschema import validate

    if isinstance(json_schema, str):
        json_schema = read_json_schema(json_schema)

    validate(obj, json_schema)
    return obj



import logging
from datetime import datetime
from math import nan
from typing import Iterator, List, Tuple

import win32evtlog
import winerror
import wmi
from typing_extensions import Self

logger = logging.getLogger(__name__)


class Resource:
    def __init__(self, wmii=None) -> None:
        if wmii:
            self.wmii = wmii
        else:
            self.wmii = wmi.WMI()

    def __iter__(self) -> Self:
        return self


class os_resources(Resource):
    def __next__(self) -> Tuple[int, int, int]:
        threads = 0
        handles = 0
        pagefaults = 0
        for proc in self.wmii.Win32_Process():
            threads += proc.ThreadCount
            handles += proc.HandleCount
            pagefaults += proc.PageFaults
        return threads, handles, pagefaults


class cpu_usage(Resource):
    def __init__(self, wmii=None, proc_name: str = "_Total") -> None:
        """proc_name is the name of the processor,
        default value mean across all processors.
        """

        Resource.__init__(self, wmii)
        self.proc_name = proc_name

        self.real0 = nan
        self.user0 = nan
        self.sys0 = nan
        self.ts0 = nan

    @staticmethod
    def get_proc_names(wmii) -> List[str]:
        return [proc.name for proc in wmii.Win32_PerfRawData_PerfOS_Processor()]

    def __next__(self) -> Tuple[float, float, float, float]:
        (p,) = self.wmii.Win32_PerfRawData_PerfOS_Processor(name=self.proc_name)

        real = int(p.PercentProcessorTime)
        user = int(p.PercentUserTime)
        sys = int(p.PercentPrivilegedTime)
        ts = int(p.Timestamp_Sys100NS)

        try:
            delta_time = ts - self.ts0
            real_ratio = 1 - (real - self.real0) / delta_time  # One Minus should not be necessary
            user_ratio = (user - self.user0) / delta_time
            sys_ratio = (sys - self.sys0) / delta_time
        except ZeroDivisionError:
            real_ratio = nan
            user_ratio = nan
            sys_ratio = nan

        self.real0 = real
        self.user0 = user
        self.sys0 = sys
        self.ts0 = ts

        return (real_ratio, user_ratio, sys_ratio, delta_time / 10000000)


class ram_usage(Resource):
    def __next__(self) -> float:
        (p,) = self.wmii.Win32_OperatingSystem()
        total = int(p.TotalVisibleMemorySize)
        free = int(p.FreePhysicalMemory)
        assert total > 0
        return 1 - (free / total)


class hdd_usage(Resource):
    def __next__(self) -> List[Tuple[str, float]]:
        ret = []
        for disk in self.wmii.Win32_LogicalDisk(DriveType=3):
            total = int(disk.Size)
            free = int(disk.FreeSpace)
            ret.append((disk.Caption, 1 - (free / total)))
        return ret


def event_logs(server: str = "localhost", source: str = "System") -> Iterator[dict]:
    """
    EventType: severity level

    EVENTLOG_ERROR_TYPE 0x0001
    EVENTLOG_WARNING_TYPE 0x0002
    EVENTLOG_INFORMATION_TYPE 0x0004
    EVENTLOG_AUDIT_SUCCESS 0x0008
    EVENTLOG_AUDIT_FAILURE 0x0010

    EventID: same for same class of messages, stringinserts are provides to template belonging to eventid

    """

    fields = {
        "Reserved",
        "RecordNumber",
        "TimeGenerated",
        "TimeWritten",
        "EventID",
        "EventType",
        "EventCategory",
        "ReservedFlags",
        "ClosingRecordNumber",
        "SourceName",
        "StringInserts",
        "Sid",
        "Data",
        "ComputerName",
    }
    fields = {
        "Reserved",
        "RecordNumber",
        "TimeGenerated",
        "TimeWritten",
        "EventID",
        "EventType",
        "EventCategory",
        "ReservedFlags",
        "ClosingRecordNumber",
        "SourceName",
        "StringInserts",
        "ComputerName",
    }

    # missing: "Sid", "Data"

    handle = win32evtlog.OpenEventLog(server, source)
    flags = win32evtlog.EVENTLOG_SEQUENTIAL_READ | win32evtlog.EVENTLOG_FORWARDS_READ

    logger.debug(f"Preparing to read {win32evtlog.GetNumberOfEventLogRecords(handle)} events")

    try:
        while True:
            events = win32evtlog.ReadEventLog(handle, flags, 0)
            if events:
                for event in events:
                    ret = {f: getattr(event, f) for f in fields}
                    # does that work for different time zones?
                    ret["TimeGenerated"] = datetime.fromtimestamp(event.TimeGenerated.timestamp())
                    ret["TimeWritten"] = datetime.fromtimestamp(event.TimeWritten.timestamp())
                    if ret["StringInserts"]:
                        ret["StringInserts"] = ", ".join(ret["StringInserts"]).replace("\r\n", "\n")
                    if not ret["StringInserts"]:
                        ret["StringInserts"] = None
                    ret["EventID"] = winerror.HRESULT_CODE(ret["EventID"])
                    # print(ret)
                    yield ret
            else:
                break
    finally:
        win32evtlog.CloseEventLog(handle)


if __name__ == "__main__":
    from argparse import ArgumentParser
    from time import sleep

    from genutility.stdio import print_line

    parser = ArgumentParser(description="Monitor OS resource usage.")
    args = parser.parse_args()

    wmii = wmi.WMI()

    for a, b, c, d in zip(cpu_usage(wmii), ram_usage(wmii), hdd_usage(wmii), os_resources(wmii)):
        print("CPU:", a)
        print("RAM:", b)
        print("HDD:", c)
        print("OS:", d)
        print_line()
        sleep(1)



import logging
import os
import sys
from shutil import get_terminal_size
from time import sleep
from typing import IO, Any, Callable, Optional, TypeVar, Union

from typing_extensions import Self

from .typing import ExceptionsType

T = TypeVar("T")
_terminal_width = get_terminal_size((80, 30)).columns


def print_line(char: str = "-", length: int = _terminal_width, file: Optional[IO] = None, flush: bool = False) -> None:
    """file=None, defaults to sys.stdout in print(). print() handles sys.stdout==None fine."""

    print(char * length, end="", file=file, flush=flush)


class PrintOnError:
    """A context manager which can be used to output data
    before an exception is printed by the python runtime.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """All arguments are passed to the `print` function."""

        self.args = args
        self.kwargs = kwargs

    def __enter__(self) -> Self:
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type:
            print(*self.args, **self.kwargs)


def safe_input(s: str, block: int = 10) -> str:
    """Makes sure ctrl-c on input() only raises KeyboardInterrupt and not EOFError+KeyboardInterrupt.
    Waits at most `block` seconds to be sure.
    """

    try:
        return input(s)
    except EOFError:
        sleep(block)  # can be interrupted by KeyboardInterrupt
        raise


def info_print(msg: Optional[str] = None, args: tuple = tuple(), exception: Optional[Exception] = None) -> None:
    # not (msg or exception) this doesn't do anything

    if exception and msg:
        logging.exception(msg % args, exc_info=exception)
        # logging.exception(msg, *args, exc_info=exception) this fails for some weird reason
    else:
        if exception:
            logging.exception("Unhandled exception", exc_info=exception)
        if msg:
            print(msg % args)


def input_ex(s: str, file_out: IO = sys.stdout, file_in: IO = sys.stdin) -> str:
    print(s, file=file_out, end="")
    file_out.flush()
    return sys.stdin.readline().rstrip("\n")


# was: raw_input_ex
def input_type(
    s: str,
    type: Optional[Callable[[str], T]] = None,
    predicate: Optional[Callable[[str], bool]] = None,
    errormsg: Optional[str] = None,
    exception: ExceptionsType = Exception,
    file: IO = sys.stdout,
) -> Union[str, T]:
    while True:
        ret = input_ex(s, file_out=file)
        if predicate and not predicate(ret):
            if errormsg:
                print(errormsg % (ret,))
            continue
        if type:
            try:
                return type(ret)
            except exception:
                if errormsg:
                    print(errormsg % (ret,))
                continue
        return ret


def confirm(msg: str, yesno: bool = True) -> bool:
    if yesno:
        s = input_type(msg + " (yes, no): ", predicate=lambda x: x.strip().lower() in {"yes", "y", "n", "no"})
        return s.strip().lower() in {"yes", "y"}
    else:
        return input_type(msg + " (anything, nothing): ", type=bool)  # bool cannot throw if given string I think...


def waitquit(msg=None, args=tuple(), exception: Optional[Exception] = None):  # fixme: ferutility.printing breaks this
    info_print(msg, args, exception)
    input("Press enter to exit...")
    sys.exit(msg or exception)


def waitcontinue(msg=None, args=tuple(), exception: Optional[Exception] = None):
    info_print(msg, args, exception)
    input("Press enter to continue...")


def errorquit(msg=None, args=tuple(), exception: Optional[Exception] = None):
    info_print(msg, args, exception)
    sys.exit(msg or exception)


def print_terminal_progress_line(out: str, file=None) -> None:
    columns, lines = os.get_terminal_size()
    if len(out) >= columns:
        out = out[: columns - 4] + "..."
    print(out, end="\r", file=file)



from typing import Any, Hashable, List

from unqlite import UnQLite


def query_by_field_eq(db: UnQLite, col: str, key: Hashable, value: Any) -> List[dict]:
    script = """
        $zCallback = function($doc) {
            return $doc[$key] == $value;
        };

        $data = db_fetch_all($col, $zCallback);
    """

    with db.vm(script) as vm:
        vm["col"] = col
        vm["key"] = key
        vm["value"] = value
        vm.execute()
        return vm["data"]


def query_by_field_intersect(db: UnQLite, col: str, key: Hashable, value: List[Any]) -> List[dict]:
    script = """
        $zCallback = function($doc) {
            return count(array_intersect($doc[$key], $value)) > 0;
        };

        $data = db_fetch_all($col, $zCallback);
    """

    with db.vm(script) as vm:
        vm["col"] = col
        vm["key"] = key
        vm["value"] = value
        vm.execute()
        return vm["data"]



import os.path
from io import SEEK_END, SEEK_SET, BufferedIOBase, RawIOBase, TextIOBase, TextIOWrapper
from mmap import mmap
from os import PathLike, fdopen
from pathlib import Path
from sys import stdout
from typing import IO, Callable, Iterable, Iterator, Optional, TypeVar, Union, overload

from typing_extensions import Self

from ._files import PathType
from .callbacks import Progress
from .iter import consume, iter_equal, resizer
from .math import PosInfInt
from .ops import logical_implication, logical_xor

Data = TypeVar("Data", str, bytes)
AnyIoT = Union[BufferedIOBase, RawIOBase, TextIOBase, mmap]  # buffered/unbuffered binary/text
BufferedIoT = Union[BufferedIOBase, TextIOBase, mmap]
BufferedBinaryIoT = Union[BufferedIOBase, mmap]
BinaryIoT = Union[RawIOBase, BufferedIOBase]
TextIoT = TextIOBase
FILE_IO_BUFFER_SIZE = 1024 * 1024


class Empty(OSError):
    pass


def _check_arguments(mode: str, encoding: Optional[str] = None) -> Optional[str]:
    is_text = "t" in mode
    is_binary = "b" in mode

    if not logical_xor(is_text, is_binary):
        raise ValueError(f"Explicit text or binary mode required: {mode}")

    if not logical_implication(is_binary, encoding is None):
        raise ValueError("Encoding is not None for binary file")

    if is_text:
        encoding = encoding or "utf-8"

    return encoding


def _stripmode(mode: str) -> str:
    return "".join(set(mode) - {"t", "b"})


def _assert_buffered_seekable_reader(fp: BufferedIoT) -> None:
    if not fp.readable or not fp.seekable:
        raise TypeError("fp need to be a buffered, readable and seekable IO type")


def read_file(
    path: PathType, mode: str = "b", encoding: Optional[str] = None, errors: Optional[str] = None
) -> Union[str, bytes]:
    """Reads and returns whole file. If content is not needed use consume_file()"""

    encoding = _check_arguments(mode, encoding)

    if "r" not in mode:
        mode = "r" + mode

    with open(path, mode, encoding=encoding, errors=errors) as fr:
        return fr.read()  # buffered


def write_file(
    data: Union[str, bytes],
    path: PathType,
    mode: str = "wb",
    encoding: Optional[str] = None,
    errors: Optional[str] = None,
) -> None:
    """Writes file."""

    encoding = _check_arguments(mode, encoding)

    with open(path, mode, encoding=encoding, errors=errors) as fw:
        fw.write(data)  # buffered


@overload
def read_or_raise(fin: BufferedBinaryIoT, size: int) -> bytes: ...


@overload
def read_or_raise(fin: TextIoT, size: int) -> str: ...


def read_or_raise(fin, size):
    """Instead of returning a result smaller than `size` like `io.read()`,
    it raises and EOFError.
    """

    data = fin.read(size)
    if len(data) == 0:  # fixme: add `and size != 0`?
        raise Empty
    elif len(data) != size:
        raise EOFError

    return data


def get_file_range(path: PathType, start: int, size: int) -> bytes:
    with open(path, "rb") as fp:
        fp.seek(start)
        return fp.read(size)  # buffered


def truncate_file(path: PathType, size: int) -> None:
    with open(path, "r+b") as fp:
        fp.truncate(size)


def wrap_text(bf, mode, encoding, errors, newline):
    if "t" in mode:
        tf = TextIOWrapper(bf, encoding, errors, newline)
        tf.mode = mode
        return tf
    return bf


def copen(
    file: Union[PathType, IO, int],
    mode: str = "rt",
    archive_file: Optional[str] = None,
    encoding: Optional[str] = None,
    errors: Optional[str] = None,
    newline: Optional[str] = None,
    compresslevel: int = 9,
    ext: Optional[str] = None,
    handle_archives: bool = True,
) -> IO:
    """Generic file open method. It supports transparent compression and improved text-mode handling.

    `file`: Can be a path, file-like or file descriptor.
    `mode`, `errors` and `newline`: see `io.open`
    `archive_file`: if a zip file is opened, this specifies the file within the archive.
    `encoding`: if `None` it defaults to "utf-8" in text-mode. It doesn't use any locale.
    `compresslevel`: 0-9, 0: no compression, 1: least, 9: highest compression.
            Only used if a compressed format is specified.
    `ext`: if `file` is a file-like object, than this can be one of {".gz", ".bz2", ".zip"}
            to enable transparent compression
    `handle_archives`: Allow transparent handling of archives. Defaults to `True`.
        If `archive_file` is not given for archives which require it, the archive file will be treated as a normal file.

    Returns a file-like object. If `file` was a fd,
            it will be closed when the file-like object is closed.
    """

    encoding = _check_arguments(mode, encoding)

    if archive_file is not None and handle_archives is False:
        raise ValueError("handle_archives must be True to allow archive_file to be used")

    if isinstance(file, (str, PathLike)):
        ext = os.path.splitext(file)[1].lower()
    elif isinstance(file, int):
        file = fdopen(file, mode, encoding=encoding, errors=errors, newline=newline)
        ext = ext and ext.lower()
    else:
        ext = ext and ext.lower()

    if handle_archives:
        if ext == ".gz":
            import gzip

            return gzip.open(file, mode, compresslevel=compresslevel, encoding=encoding, errors=errors, newline=newline)

        elif ext == ".bz2":
            import bz2

            return bz2.open(file, mode, compresslevel=compresslevel, encoding=encoding, errors=errors, newline=newline)

        elif ext == ".zip" and archive_file:
            from zipfile import ZipFile

            newmode = _stripmode(mode)

            with ZipFile(
                file, newmode
            ) as zf:  # note: even if the outer zip file is closed, the inner file can still be read apparently
                bf = zf.open(archive_file, newmode, force_zip64=True)

            return wrap_text(bf, mode, encoding, errors, newline)

    if isinstance(file, TextIOBase):
        return file

    if hasattr(file, "read") or hasattr(file, "write"):  # file should be in binary mode here
        return wrap_text(file, mode, encoding, errors, newline)

    if isinstance(file, Path):
        return file.open(mode, encoding=encoding, errors=errors, newline=newline)
    else:
        return open(file, mode, encoding=encoding, errors=errors, newline=newline)


class OpenFileAndDeleteOnError:
    """Context manager which opens a file using the same arguments as `open`,
    but deletes the file in case an exception occurs after opening.
    """

    def __init__(
        self,
        file: PathType,
        mode: str = "rt",
        encoding: Optional[str] = None,
        errors: Optional[str] = None,
        newline: Optional[str] = None,
        compresslevel: int = 9,
    ) -> None:
        encoding = _check_arguments(mode, encoding)

        self.file = file
        self.mode = mode
        self.encoding = encoding
        self.errors = errors
        self.newline = newline
        self.compresslevel = compresslevel
        self.fp: Optional[IO] = None

    def __enter__(self) -> IO:
        self.fp = copen(self.file, self.mode, None, self.encoding, self.errors, self.newline, self.compresslevel)
        return self.fp

    def __exit__(self, exc_type, exc_value, traceback):
        self.fp.close()
        if exc_type:
            # fixme: race condition
            # use https://stackoverflow.com/a/3594593 and `ReOpenFile` on Windows
            # or SetFileInformationByHandle
            # see also: https://nullprogram.com/blog/2016/08/07/
            os.remove(self.file)


class OptionalWriteOnlyFile:
    def __init__(
        self,
        path: Optional[PathType] = None,
        mode: str = "xb",
        encoding: Optional[str] = None,
        errors: Optional[str] = None,
        newline: Optional[str] = None,
        compresslevel: int = 9,
    ) -> None:
        if path:
            self.fp: Optional[IO] = copen(
                path, mode, encoding=encoding, errors=errors, newline=newline, compresslevel=compresslevel
            )
        else:
            self.fp = None

    def __enter__(self):
        if self.fp:
            return self.fp
        else:
            return self

    def close(self) -> None:
        if self.fp:
            self.fp.close()

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def write(self, data) -> Optional[int]:
        pass

    def seek(self, offset: int, whence: int = 0) -> int:
        pass


class StdoutFile:
    def __init__(
        self,
        path: Optional[PathType] = None,
        mode: str = "xb",
        encoding: Optional[str] = None,
        errors: Optional[str] = None,
        newline: Optional[str] = None,
        compresslevel: int = 9,
    ) -> None:
        encoding = _check_arguments(mode, encoding)

        if path:
            self.fp = copen(path, mode, encoding=encoding, errors=errors, newline=newline, compresslevel=compresslevel)
            self.doclose = True
        else:
            self.doclose = False
            if "b" in mode:
                self.fp = stdout.buffer
            elif "t" in mode:
                self.fp = stdout
            else:
                raise ValueError(f"Explicit text or binary mode required: {mode}")

    def __enter__(self):
        return self.fp

    def close(self) -> None:
        if self.doclose:
            self.fp.close()

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


class PathOrBinaryIO:
    def __init__(self, fname: Union[PathType, IO[bytes]], mode: str = "rb", close: bool = False) -> None:
        if isinstance(fname, (RawIOBase, BufferedIOBase)):
            self.doclose: bool = close
            self.fp: IO[bytes] = fname
        else:
            self.doclose = True
            self.fp = copen(fname, mode)

    def __enter__(self) -> IO[bytes]:
        return self.fp

    def close(self) -> None:
        if self.doclose:
            self.fp.close()

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


class PathOrTextIO:
    def __init__(
        self,
        fname: Union[PathType, IO[str]],
        mode: str = "rt",
        encoding: str = "utf-8",
        errors: str = "strict",
        newline: Optional[str] = None,
        close: bool = False,
    ) -> None:
        if isinstance(fname, TextIOBase):
            self.doclose: bool = close
            self.fp: IO[str] = fname
        else:
            self.doclose = True
            self.fp = copen(fname, mode, encoding=encoding, errors=errors, newline=newline)

    def __enter__(self) -> IO[str]:
        return self.fp

    def close(self) -> None:
        if self.doclose:
            self.fp.close()

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


class LastLineFile:
    chunk_size = 1024 * 4
    nl = "\n"

    def __init__(self, path, mode="rt+"):
        self.f = open(path, mode)  # buffered
        self.ll_pos = None

    def __enter__(self) -> Self:
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        self.f.close()

    def _seek_to_last_line(self):
        if self.ll_pos is None:
            self.ll_pos = self._get_last_line_pos()

        self.f.seek(self.ll_pos, SEEK_SET)

    def _get_last_line_pos(self):
        self.f.seek(0, SEEK_END)
        pos = self.f.tell()
        ret_pos = 0
        while pos > 0:
            seek_next = max(pos - self.chunk_size, 0)
            self.f.seek(seek_next, SEEK_SET)
            chunk = self.f.read(pos - seek_next)
            fpos = chunk.rfind("\n")
            if fpos != -1:
                ret_pos = seek_next + fpos + 1
                if ret_pos != pos:
                    break
            pos = seek_next
        return ret_pos

    def _validate(self, s):
        if self.nl in s:
            raise ValueError("Newline in input")

    def read(self):
        self._seek_to_last_line()
        return self.f.read()

    def newline(self, s):
        self.f.seek(0, SEEK_END)
        ret = self.f.write(self.nl)
        self.ll_pos = self.f.tell()
        return ret

    def replace(self, s):
        self._validate(s)
        self._seek_to_last_line()
        ret = self.f.write(s)
        self.f.truncate()
        return ret


class Tell:
    def __init__(self, fp: BufferedIoT) -> None:
        try:
            assert not fp.seekable()
        except AttributeError:
            pass

        self._fp = fp
        self._pos = 0

    # wrapped

    def read(
        self, size=None
    ):  # the docs for `io.BufferedIOBase` say that size=-1, but that's not true for `HTTPResponse`
        ret = self._fp.read(size)
        self._pos += len(ret)
        return ret

    def read1(self, n=-1):
        ret = self._fp.read1(n)
        self._pos += len(ret)
        return ret

    def readall(self):
        ret = self._fp.readall()
        self._pos += len(ret)
        return ret

    def readline(self, limit=-1):
        ret = self._fp.readline(limit)
        self._pos += len(ret)
        return ret

    def readlines(self, hint=-1):
        raise NotImplementedError
        return self._fp.readlines(hint)

    def readinto(self, b):
        ret = self._fp.readinto(b)
        self._pos += ret
        return ret

    def readinto1(self, b):
        ret = self._fp.readinto1(b)
        self._pos += ret
        return ret

    def write(self, b):  # fixme: should this be added to _pos?
        return self._fp.write(b)

    def writelines(self, lines):  # fixme: should this be added to _pos?
        return self._fp.writelines(lines)

    def seek(self, offset, whence=SEEK_SET):
        raise OSError("Stream is not seekable")

    def tell(self):
        return self._pos

    # redirected

    def close(self):
        return self._fp.close()

    def detach(self):
        return self._fp.detach()

    def fileno(self):
        return self._fp.fileno()

    def flush(self):
        return self._fp.flush()

    def isatty(self):
        return self._fp.isatty()

    def readable(self):
        return self._fp.readable()

    def seekable(self):
        return self._fp.seekable()

    def truncate(self, size=None):
        return self._fp.truncate(size)

    def writable(self):
        return self._fp.writable()


from collections import deque
from urllib import response


class BufferedTell(response.addinfourl):  # fixme: untested!!!
    def __init__(self, filesize):
        # response.addinfourl.__init__(self, *args, **kwargs)
        self._pos = 0
        self._read = 0
        self.filesize = filesize
        self.buf = deque(maxlen=1024)

    def read(self, num=-1):
        if num == -1:
            num = self.filesize - self._pos
            print("num=-1")
        if self._pos == self._read:
            self._pos += num
            self._read += num
            data = response.addinfourl.read(num)
            self.buf.extend(data)
            return data
        elif self._pos < self._read:
            delta = self._read - self._pos
            self._pos += num
            if num <= delta:
                data = b"".join(self.buf.pop() for i in range(delta))
                self.buf.extend(data[::-1])
                return data[-1 : -1 - num : -1]
            else:
                datab = b"".join(self.buf.pop() for i in range(delta))
                datab = datab[::-1]
                self.buf.extend(datab)
                dataf = response.addinfourl.read(num - delta)
                self.buf.extend(dataf)
                return datab + dataf
        else:
            # return b""
            raise Exception("Unbuffered read")

    def tell(self):
        return self._pos

    def seek(self, offset, whence=0):
        if whence == 0:
            self._pos = offset
        elif whence == 1:
            self._pos += offset
        elif whence == 2:
            self._pos = self.filesize - offset
        else:
            raise ValueError("Unknown whence")


def copyfilelike(
    fin: AnyIoT,
    fout: BufferedIoT,
    amount: Optional[int] = None,
    buffer: int = FILE_IO_BUFFER_SIZE,
    report: Optional[Callable[[int, int], None]] = None,
) -> int:
    """Read data from `fin` in chunks of size `buffer` (or less) and write them to `fout`.
    Optionally limit the amount of data to `amount`. `report` can be a callable which receives
    the total number of bytes copied and bytes remaining.
    Input and Output streams but both be either binary or both text.
    see `shutil.copyfileobj`
    """

    # todo: have different input and output buffers

    _amount = amount or PosInfInt

    copied = 0
    while _amount > 0:
        if report:
            report(copied, _amount + copied)

        data = fin.read(min(buffer, _amount))  # buffered or unbuffered

        if not data:
            break
        fout.write(data)
        _amount -= len(data)
        copied += len(data)
    return copied


def simple_file_iter(fr: IO[Data], chunk_size: int = FILE_IO_BUFFER_SIZE) -> Iterator[Data]:
    """Iterate file-like object `fr` and yield chunks of size `chunk_size` (or less in unbuffered mode)."""

    assert isinstance(chunk_size, int), "chunk_size needs to be an integer"

    # return iter(partial(fr.read, chunk_size), "") sentinel cannot be defined well
    while True:
        data = fr.read(chunk_size)
        if data:
            yield data
        else:
            break


def reversed_file_iter(fp: BufferedBinaryIoT, chunk_size: int = FILE_IO_BUFFER_SIZE) -> Iterator[bytes]:
    """Generate blocks of file's contents in reverse order."""

    _assert_buffered_seekable_reader(fp)

    fp.seek(0, SEEK_END)
    here = fp.tell()
    while here > 0:
        delta = min(chunk_size, here)
        here -= delta
        fp.seek(here, SEEK_SET)
        yield fp.read(delta)


def limited_file_iter(fr: IO[Data], amount: int, chunk_size: int = FILE_IO_BUFFER_SIZE) -> Iterator[Data]:
    """Iterate file-like object `fr` and yield chunks of size `chunk_size`.
    Limit output to `amount` bytes.
    """

    while amount > 0:
        data = fr.read(min(chunk_size, amount))
        if not data:
            break
        yield data
        amount -= len(data)


def iterfilelike(
    fr: IO[Data], start: int = 0, amount: Optional[int] = None, chunk_size: int = FILE_IO_BUFFER_SIZE
) -> Iterable[Data]:
    """Iterate file-like object `fr` and yield chunks of size `chunk_size`.
    Starts reading at `start` and optionally limit output to `amount` bytes.
    """

    if start:
        fr.seek(start)

    if amount is None:
        return simple_file_iter(fr, chunk_size)
    else:
        return limited_file_iter(fr, amount, chunk_size)


def blockfileiter(
    path: PathType,
    mode: str = "rb",
    buffering: int = -1,
    encoding: Optional[str] = None,
    errors: Optional[str] = None,
    start: int = 0,
    amount: Optional[int] = None,
    chunk_size: int = FILE_IO_BUFFER_SIZE,
    opener=None,
) -> Iterator[Union[str, bytes]]:
    """Iterate over file at `path` and yield chunks of size `chunk_size` (or less in unbuffered mode).
    Optionally limit output to `amount` bytes.
    """

    encoding = _check_arguments(mode, encoding)

    with open(path, mode, buffering, encoding, errors=errors, opener=opener) as fr:
        yield from iterfilelike(fr, start, amount, chunk_size)


def blockfilesiter(paths: Iterable[PathType], chunk_size: int = FILE_IO_BUFFER_SIZE) -> Iterator[bytes]:
    """Iterate over a list of files and return their contents in a
    continuous stream of chunks of size `chunk_size`.
    That means one chunk can span multiple files.
    """

    chunk = b""

    for path in paths:
        with open(path, "rb") as fr:
            if len(chunk) != 0:
                data = fr.read(chunk_size - len(chunk))  # buffered
                chunk += data  # fixme: probably slow for lots of small files

            if len(chunk) == chunk_size:
                yield chunk
                chunk = b""

            if len(chunk) == 0:
                for data in iterfilelike(fr, chunk_size=chunk_size):
                    if len(data) == chunk_size:
                        yield data
                    else:  # must be the last chunk in buffered mode
                        chunk = data

    if chunk:
        yield chunk


def bufferedfileiter(
    path: PathType,
    chunk_size: int,
    mode: str = "rb",
    buffering: int = -1,
    encoding: Optional[str] = None,
    errors: Optional[str] = None,
    start: int = 0,
    amount: Optional[int] = None,
    buffer_size: int = FILE_IO_BUFFER_SIZE,
) -> Iterable[Union[str, bytes]]:
    """Iterate over file at `path` reading chunks of size `buffer_size` at a time.
    Yields data chunks of `chunk_size` and optionally limits output to `amount` bytes.
    """

    return resizer(blockfileiter(path, mode, buffering, encoding, errors, start, amount, buffer_size), chunk_size)


def byte_out(path: PathType, buffer_size: int = FILE_IO_BUFFER_SIZE) -> Iterable[bytes]:
    """Reads file at `path` byte by byte, using a read buffer of size `buffer_size`."""

    return resizer(blockfileiter(path, "rb", chunk_size=buffer_size), 1)


def consume_file(filename: PathType, buffer_size: int = FILE_IO_BUFFER_SIZE) -> None:
    """reads whole file but ignores content"""

    consume(blockfileiter(filename, "rb", chunk_size=buffer_size))


# was: same_files, textfile_equal: equal_files(*paths, mode="rt")
def equal_files(
    *paths: PathType,
    mode: str = "rb",
    encoding: Optional[str] = None,
    errors: Optional[str] = None,
    amount: Optional[int] = None,
    chunk_size: int = FILE_IO_BUFFER_SIZE,
) -> bool:
    """Check if files at `*paths` are equal. Chunks of size `chunk_size` are read at a time.
    Data can be optionally limited to `amount`.
    """

    # blockfileiter must be used in buffered mode, since the chunks can be of different size
    its = tuple(
        blockfileiter(path, mode, encoding=encoding, errors=errors, amount=amount, chunk_size=chunk_size)
        for path in paths
    )
    return iter_equal(*its)


def is_all_byte(fr: BinaryIoT, thebyte: bytes = b"\0", chunk_size: int = FILE_IO_BUFFER_SIZE) -> bool:
    """Test if file-like `fr` consists only of `thebyte` bytes."""

    assert isinstance(thebyte, bytes)

    thebyte = thebyte * chunk_size
    for data in simple_file_iter(fr, chunk_size):
        if data != thebyte[: len(data)]:
            return False
    return True


def iter_lines(
    path: PathType,
    encoding: str = "utf-8",
    errors: str = "strict",
    newline: Optional[str] = None,
    progress: Optional[Progress] = None,
) -> Iterator[str]:
    """Yield lines from text file. Handles compressed files as well.
    Use `progress` for progress reporting.
    """

    with copen(path, "rt", encoding=encoding, errors=errors, newline=newline) as fr:
        if progress is not None:
            prev_pos = 0

            fr.seek(0, os.SEEK_END)
            total = fr.tell()
            fr.seek(0, os.SEEK_SET)

            with progress.task(total=total) as task:
                while True:  # cannot use `for line in fr` here
                    line = fr.readline()
                    if not line:
                        break
                    cur_pos = fr.tell()
                    task.advance(cur_pos - prev_pos)
                    yield line
                    prev_pos = cur_pos
        else:
            for line in fr:
                yield line


def iter_stripped(
    path: PathType, encoding: str = "utf-8", errors: str = "strict", newline: Optional[str] = None
) -> Iterator[str]:
    """Yield lines from text file with trailing newlines stripped.
    Handles compressed files as well.
    """

    with copen(path, "rt", encoding=encoding, errors=errors, newline=newline) as fr:
        for line in fr:
            yield line.rstrip("\r\n")


# is this still needed?
def file_byte_reader(
    filename: PathType, inputblocksize: int, outputblocksize: int, DEBUG: bool = True
) -> Iterator[bytes]:
    assert (inputblocksize % outputblocksize == 0) or (
        outputblocksize % inputblocksize == 0
    ), "Neither input nor output size is a multiple of the other"

    bytes_yielded = 0
    bytes = bytearray(max(inputblocksize, outputblocksize))
    bytes_used = 0
    for read in blockfileiter(filename, chunk_size=inputblocksize):
        bytes[bytes_used : bytes_used + len(read)] = read
        # print("read {} bytes to pos [{}:{}]".format(len(read),bytes_used,bytes_used+len(read)))
        bytes_used += len(read)
        pos = 0
        while bytes_used >= outputblocksize:
            yield bytes[pos : pos + outputblocksize]
            bytes_yielded += outputblocksize
            # print("yielded {} bytes from pos [{}:{}]".format(outputblocksize,pos,pos+outputblocksize))
            pos += outputblocksize
            bytes_used -= outputblocksize
    if DEBUG:
        import os.path

        filesize = os.path.getsize(filename)
        if filesize != bytes_yielded:
            print(f"{bytes_yielded} bytes yielded, filesize: {filesize}")



import logging
import random
import sys
from collections import deque
from functools import partial
from heapq import heappop, heappush, heapreplace
from itertools import chain, combinations, count, groupby, islice, product, repeat, starmap, tee, zip_longest
from operator import add, itemgetter
from random import randrange
from time import sleep, time
from types import GeneratorType
from typing import (
    Any,
    Callable,
    Collection,
    Container,
    Dict,
    Generic,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Set,
    TextIO,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from .exceptions import EmptyIterable, IteratorExhausted
from .ops import operator_in
from .typing import ExceptionsType, Orderable

OrderableT = TypeVar("OrderableT", bound=Orderable)

T = TypeVar("T")
U = TypeVar("U")
V = TypeVar("V")

logger = logging.getLogger(__name__)


def one(it: Iterator[T]) -> T:
    try:
        ret = next(it)
    except StopIteration:
        raise EmptyIterable()
    try:
        next(it)
    except StopIteration:
        pass
    else:
        raise ValueError("Iterator yielded more than one value")

    return ret


def iterrandrange(a: int, b: int) -> Iterator[int]:
    """Yields a stream of non-cryptographic random numbers between `a` (inclusive) and `b` (exclusive)."""

    while True:
        yield randrange(a, b)  # nosec


def repeatfunc(func: Callable[..., T], times: Optional[int] = None, *args: Any) -> Iterator[T]:
    """Repeat calls to func with specified arguments.
    Example:  repeatfunc(random.random)
    See: https://docs.python.org/3/library/itertools.html#itertools-recipes
    """

    if times is None:
        return starmap(func, repeat(args))
    return starmap(func, repeat(args, times))


def _lstr(it: Iterable, length: Optional[int] = None) -> Tuple[Optional[int], str]:
    if length is None:
        try:
            length = len(it)  # type: ignore[arg-type]
        except TypeError:
            pass

    if length is not None:
        lstr = f"/{length}"
    else:
        lstr = ""

    return length, lstr


def progress(
    it: Union[Iterable, Collection],
    length: Optional[int] = None,
    refresh: float = 1,
    end: str = "\r",
    file: Optional[TextIO] = None,
    extra_info_callback: Optional[Callable[[int, Optional[int]], str]] = None,
    disable: Optional[bool] = None,
    delta: Optional[int] = 1,
) -> Iterator:
    """Wraps an iterable `it` to periodically print the progress every `refresh` seconds.
    `length` is the total size of `it`. `len(it)` is used to get the size if set to `None`.
    `refresh` updates the printed output every `refresh` seconds.
    `file` write to file object, defaults to stdout.
    `delta` specifies the value which is added to the counter every iteration.
            `None` uses `len(elm)` to determine the value. Defaults to 1.
    """

    """ todo: `from operator import length_hint`
        Use `length_hint(it)` for a guess which might be better than nothing.
        Can return 0.
    """

    file = file or sys.stdout

    if delta is not None and delta < 1:
        raise ValueError(f"`delta` must be a integer larger than 1 or `None`, not '{delta}'")

    if disable is None and hasattr(file, "isatty") and not file.isatty():
        disable = True

    if disable:
        yield from it
        return

    length, lstr = _lstr(it, length)
    extra = ""
    last = start = time()
    total = 0

    if extra_info_callback:
        extra = extra_info_callback(0, length)
        extra = f" [{extra}]"
    print(f"0{lstr}, running for 0 seconds{extra}.", end="\r", file=file)

    try:
        for elm in it:
            yield elm
            if delta is None:
                total += len(elm)
            else:
                total += delta
            current = time()
            if current - last > refresh:
                last = current
                duration = current - start
                if extra_info_callback:
                    extra = extra_info_callback(total, length)
                    extra = f" [{extra}]"
                print(
                    f"{total}{lstr}, running for {int(duration)} seconds ({total/duration:0.2e}/s){extra}.",
                    end="\r",
                    file=file,
                )
    except KeyboardInterrupt:
        duration = last - start
        if duration > 0:
            print(
                f"Unsafely aborted after reading {total}{lstr} in {int(duration)} seconds ({total/duration:0.2e}/s).",
                end=end,
                file=file,
            )
        else:
            print(f"Unsafely aborted after reading {total}{lstr} in {int(duration)} seconds.", end=end, file=file)
        raise
    else:
        duration = last - start
        if duration > 0:
            print(f"Finished {total} in {int(duration)} seconds ({total/duration:0.2e}/s).", end=end, file=file)
        else:
            print(f"Finished {total} in {int(duration)} seconds.", end=end, file=file)


def progressdata(
    it: Union[Iterable, Collection],
    length: Optional[int] = None,
    refresh: float = 1,
    end: str = "\r",
    file: Optional[TextIO] = None,
    extra_info_callback: Optional[Callable[[int, Optional[int]], str]] = None,
    disable: Optional[bool] = None,
) -> Iterable:
    return progress(it, length, refresh, end, file, extra_info_callback, disable, None)


class Progress:
    """Cannot create a decorator to do this to generic generators,
    because attributes cannot be set on generator objects.
    """

    def __init__(self, obj: Collection, *args, **kwargs) -> None:
        self.obj = obj
        self.args = args
        self.kwargs = kwargs

    def __iter__(self) -> Iterator:
        return progress(self.obj, len(self.obj), *self.args, **self.kwargs)

    def __len__(self) -> int:
        """Can raise a TypeError if underlying object does not support `len()`."""

        return len(self.obj)


def count_distinct(it: Iterable[Any]) -> int:
    return len(set(it))


def reversedzip(*its):
    return zip(*map(reversed, its))


def flatten(it: Union[tuple, list]) -> Iterator:
    for elm in it:
        if isinstance(elm, (tuple, list)):
            yield from flatten(elm)
        else:
            yield elm


def diffsgn(a: OrderableT, b: OrderableT) -> int:
    """Sign of the finite derivative.
    Equal to `-cmp(a, b)` or `cmp(b, a).
    """

    return (a < b) - (a > b)


def asc_peaks(iterable: Iterable[OrderableT]) -> Iterator[OrderableT]:
    """Yields ascending peaks from `iterable`."""

    it = iter(iterable)
    try:
        cur = next(it)
    except StopIteration:
        return

    for x in it:
        if x < cur:
            yield cur
        cur = x
    yield cur


def extrema(
    iterable: Iterable[OrderableT],
    first: Set[int] = frozenset((1, -1)),
    last: Set[int] = frozenset((1, -1)),
    derivatives: Set[int] = frozenset((1, -1)),
) -> Iterator[OrderableT]:
    it = iter(iterable)
    try:
        old = next(it)
    except StopIteration:
        return
    try:
        old_2 = next(it)
        old_d = diffsgn(old, old_2)
    except StopIteration:
        yield old  # fixme: should this be returned?
        return

    if old_d in first:
        yield old

    old = old_2

    for new in it:
        new_d = diffsgn(old, new)
        if new_d != old_d and old_d in derivatives:
            yield old
        old = new
        old_d = new_d

    if old_d in last:
        yield old


def peaks(it: Iterable[OrderableT]) -> Iterator[OrderableT]:
    """Yields peaks of `it`."""

    return extrema(it, {-1}, {1}, {1})


def valleys(it: Iterable[OrderableT]) -> Iterator[OrderableT]:
    """Yields valleys of `it`."""

    return extrema(it, {1}, {-1}, {-1})


def empty() -> Iterator:
    """An empty iterator. Other methods include:
    iter(())
    yield from ()
    """

    return
    yield  # type: ignore # pylint: disable=unreachable


def lastdefault(it: Iterable[T], default: Optional[T] = None) -> Optional[T]:
    """Returns the last element of iterable `it`. It discards all other values.
    This method is about 1/2 times as fast as `last()` for large iterables.
    """

    ret = default
    for i in it:
        ret = i
    return ret


def last(it: Iterable[T]) -> T:
    """Returns the last element of iterable `it`. It discards all other values.
    This method is about 2 times as fast as `lastdefault()` for large iterables.
    """

    try:
        return deque(it, 1).pop()
    except IndexError:
        raise EmptyIterable("Empty iterable") from None


def batch(it: Iterable[Any], n: int, func: Optional[Callable] = None) -> Iterator[Any]:
    # return type cannot be more specific because of filter()

    """Batches iterable `it` into batches of size `n`.
    Optionally post-processes batch with `func`.
    """

    it = iter(it)
    while True:
        chunk_it = islice(it, n)
        try:
            first_el = next(chunk_it)
        except StopIteration:
            return

        batch_iter = chain((first_el,), chunk_it)

        if func:
            batch_iter = func(batch_iter)

        yield batch_iter


def advance(it: Iterable[T], n: int) -> None:
    """Advances the iterable `it` n steps."""

    """
    for i in range(n):
        next(it, None)
    """
    next(islice(it, n, n), None)


def filternone(it: Iterable) -> Iterator:
    """Removes all `None` values from `it`."""

    return (i for i in it if i is not None)


def all_equal(it: Iterable) -> bool:
    """Returns `True` if all elements of `it` are equal to each other.

    Much faster than:
            return len(set(it)) in (0, 1)
    and also faster than:
            try:
                    first = next(it)
            except StopIteration:
                    return True
            return all(first == rest for rest in it)
    for both extreme cases: all equal and none equal
    """

    g = groupby(it)
    return next(g, True) and not next(g, False)  # type: ignore


def pairwise(it: Iterable[T]) -> Iterator[Tuple[T, T]]:
    """Return two elements of iterable `it` at a time,
    but only advances the iterable by 1 each time.
    pairwise((a, b, c)) -> (a, b), (b, c)
    """

    a, b = tee(it, 2)
    next(b, None)
    return zip(a, b)


def x_wise(it: Iterable, size: int) -> Iterator:
    """Returns `size` elements of iterable `it`, but only advances the iterable by 1 each time.
    ([1,2,3,4,5], 2) -> [1,2], [2,3], [3,4], [4,5]
    ([1,2,3,4,5], 3) -> [1,2,3], [2,3,4], [3,4,5]
    ([1,2,3,4,5], 4) -> [1,2,3,4], [2,3,4,5]
    and so only
    """

    copies = tee(it, size)
    for copy, pos in zip(copies, range(size)):
        advance(copy, pos)
    return zip(*copies)


triples = partial(x_wise, size=3)


def findfirst(
    func: Callable[[T], bool], it: Iterable[T], default: Tuple[Optional[int], Optional[T]] = (None, None)
) -> Tuple[Optional[int], Optional[T]]:
    """Find the first element of iterable `it` where `func(elm)` evaluates to True.
    Return `default` if not such element was found.
    """

    for i, x in enumerate(it):
        if func(x):
            return i, x

    return default


def is_empty(it: Iterator) -> bool:
    """Returns True if the iterator `it` is already fully consumed.
    If not, it will be advanced.
    """

    try:
        next(it)
        return False
    except StopIteration:
        return True


def consume(it: Iterable) -> None:
    """Consumes the iterable `it` completely."""

    deque(it, maxlen=0)


def resizer(
    it: Iterable[Sequence[T]], size: int, pad: bool = False, pad_elm: Optional[T] = None
) -> Iterable[List[Optional[T]]]:
    """Cuts the input iterable `it` consisting of variable length slices
    into slices of length `size`. If pad is True, the last slice
    will be padded to length `size` with `pad_elm`.

    resizer(("asd", "qw", "e"), 4) -> (['a', 's', 'd', 'q'], ['w', 'e'])
    resizer(("asd", "qw", "e"), 4, True, "x") -> (['a', 's', 'd', 'q'], ['w', 'e', 'x', 'x'])
    """

    if size <= 0:
        raise ValueError("size must be larger than 0")

    it = iter(it)
    buf: Sequence[T] = []
    pos = 0
    try:
        while True:
            to_read = size
            out: List[Optional[T]] = []
            while to_read > 0:
                if len(buf) == pos:
                    buf = next(it)
                    pos = 0
                tmp = buf[pos : pos + to_read]
                out.extend(tmp)
                to_read -= len(tmp)
                pos += len(tmp)
            yield out
    except StopIteration:
        if out:
            while pos < len(buf):
                logger.error("this can never happen, can it?")
                tmp = buf[pos : pos + to_read]
                out.extend(tmp)
                to_read -= len(tmp)
                pos += to_read
            if pad:
                yield out + [pad_elm] * (size - len(out))
            else:
                yield out


def switch(it: Iterable[Tuple[T, U]]) -> Iterator[Tuple[U, T]]:
    """Swaps the elements in iterable of pairs `it`.
    ((3, 6), (7, 4), (1, 9)) -> (6, 3), (4, 7), (9, 1)
    """

    return ((j, i) for i, j in it)


def switched_enumerate(it: Iterable[T]) -> Iterator[Tuple[T, int]]:
    """Same as `enumerate()` except that order of the output pair is switched."""

    return switch(enumerate(it))


def multi_join(it_a: Iterable[T], it_b: Iterable[U], join_func: Callable[[T, U], V] = add) -> Iterator[V]:
    """Calls `join_func` on the cross product of `it_a` and `it_b`.
    multi_join((1, 2), (3, 4), lambda x, y: str(x)+str(y)) -> ('13', '23', '14', '24')
    """

    return (join_func(a, b) for b in it_b for a in it_a)


def iter_except(
    iterator: Iterator[T],
    exception_callbacks: Dict[Type, Callable[[Iterator[T], Exception], bool]],
    return_on_exception: bool = False,
) -> Iterator[T]:
    """Calls callbacks when exceptions are raised in `iterator`. Does not work for Generators.
    The type key in the `exception_callbacks` dict must be the same type as the exception which
    should be caught, not a subtype.
    """

    if not return_on_exception and isinstance(iterator, GeneratorType):
        raise TypeError("iterator cannot be a generator")

    while True:
        try:
            yield next(iterator)
        except StopIteration:
            return
        except Exception as e:
            try:
                if exception_callbacks[type(e)](
                    iterator, e
                ):  # use if x. and not if not x, so None doesn't raise, if user forgets to returns correct value
                    raise
            except KeyError:
                raise e
            if return_on_exception:
                return


def list_except(it: Iterable[T], catch: ExceptionsType = Exception) -> Tuple[Optional[Exception], List[T]]:
    """Same as `list()` except in the case of an exception, the partial list which was collected
    so far is returned. `catch` specifies which exceptions are caught. It can be an exception
    or a tuple of exceptions.
    """

    ret = []  # List[T]
    exc: Optional[Exception] = None

    try:
        for i in it:
            ret.append(i)
    except catch as e:
        exc = e

    return exc, ret


def iter_except_ignore(
    iterator: Iterator[T], exceptions: ExceptionsType, logger: logging.Logger = logger
) -> Iterator[T]:
    """Ignores `exceptions` raised in `iterator`. Does not work for Generators."""

    if isinstance(iterator, GeneratorType):
        raise TypeError("iterator cannot be a generator")

    while True:
        try:
            yield next(iterator)
        except StopIteration:
            break
        except exceptions as e:
            logger.warning("Error in iterable: %s", str(e))


def decompress(selectors: Iterable[bool], data: Iterator[T], default: Optional[T] = None) -> Iterator[Optional[T]]:
    """Basically the opposite of `itertools.compress`.
    decompress([True, False, True], iter(["A", "B"])) -> ("A", None, "B")
    """

    try:
        for s in selectors:
            if s:
                yield next(data)
            else:
                yield default
    except StopIteration:  # exception converted because StopIterations raised in generators will cause RuntimeErrors
        raise IteratorExhausted


def first_not_none(it: Iterable[T], default: Optional[T] = None) -> Optional[T]:
    """Returns the first element of `it` which is not None."""

    return next(filternone(it), default)


def range_count(start: int = 0, stop: Optional[int] = None, step: int = 1) -> Iterable[int]:
    """Similar to `range`, except it optionally allows for an infinite range."""

    if stop:
        return range(start, stop, step)
    else:
        return count(start, step)


def product_range_repeat(depth: int, args: Tuple) -> Iterator:
    """Returns coordinate or coordinate slices useful for iteration over space.
    Can replace higher dimensional for-in-range loops.
    """

    return product(tuple(range(*args)), repeat=depth)


def powerset(sit: Collection[T]) -> Iterator[Tuple[T, ...]]:
    """Returns the powerset of a sized iterable.
    The powerset of an empty set is a set including the empty set.

    powerset([1,2,3]) --> (), (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)
    """

    return chain.from_iterable(combinations(sit, r) for r in range(len(sit) + 1))


# had cmp=b"\0" default before
def all_equal_to(it: Iterable[T], cmp: T) -> bool:
    """Test if every item of iterable `it` is equal to `cmp`."""

    return all(elm == cmp for elm in it)


def iter_equal(*its: Iterable) -> bool:
    """Test if the contents of all iterables `its` is the same."""

    return all(all_equal(it_elements) for it_elements in zip_longest(*its))


def iter_different(it_a: Iterable, it_b: Iterable) -> bool:
    """Test if `it_a` and `it_b` yield different elements."""

    return any(a != b for a, b in zip_longest(it_a, it_b))


def every_n(it: Iterable[T], n: int, pos: int = 0) -> Iterator[T]:
    """Yields every `n`-th element from iterable `it` and starts at `pos`."""

    if pos > 0:
        advance(it, pos)

    for i, val in enumerate(it):
        if i % n == 0:
            yield val


def split(it: Iterable[T], size: int) -> Sequence[Iterable[T]]:
    """memory consumption should be proportional to `size` when output iterators
    are used in parallel loops (like `zip`), else might be higher
    """

    copies = tee(it, size)
    return tuple(every_n(it, size, pos) for it, pos in zip(copies, range(size)))


def no_dupes(*its: Iterable[T]) -> Iterator[T]:
    """Merge `its` and removes all duplicates while preserving order."""

    seen: Set[T] = set()

    for it in its:
        for x in it:
            if x not in seen:
                yield x
                seen.add(x)


def retrier(
    waittime: float,
    attempts: int = -1,
    multiplier: float = 1,
    jitter: float = 0,
    max_wait: Optional[float] = None,
    jitter_dist: str = "uniform",
    waitfunc: Callable[[float], Any] = sleep,
) -> Iterator[int]:
    """Iterator which yields after predefined amounts of time.
    Supports `multiplier` to implements linear or exponential backoff.
    `jitter` adds noise to the average wait time, distributed according to `jitter_dist`.
    Supported values for `jitter_dist` are `uniform` and `normal`.
    """

    if jitter_dist == "uniform":

        def rand(waittime, jitter):
            return waittime + random.uniform(-jitter, jitter)  # nosec

    elif jitter_dist == "normal":

        def rand(waittime, jitter):
            return random.normalvariate(waittime, jitter)

    else:
        raise ValueError("Unsupported jitter_dist")

    if attempts == 0:
        return

    for i in count():
        yield i

        attempts -= 1

        if attempts != 0:
            if max_wait and waittime > max_wait:
                jitter *= max_wait / waittime
                waittime = max_wait
                multiplier = 1

            do_wait = max(0, rand(waittime, jitter))
            waitfunc(do_wait)
        else:
            break

        waittime *= multiplier
        jitter *= multiplier


def collapse_any(it: Iterable[T], col_set: Set[T]) -> Iterator[T]:
    """Removes consecutive elements from iterable `it` if element is in `col_set`.
    ([1,1,2,2,3,3,4,4], {1,2}) -> 1, 2, 3, 3, 4, 4
    """

    assert isinstance(col_set, set)

    for key, g in groupby(it):
        if key in col_set:
            yield key
        else:
            yield from g


def collapse_all(it: Iterable[T], col_set: Set[T], replacement: T) -> Iterator[T]:
    """Removes consecutive elements from iterable `it` if element is in `col_set`
    and replace `replacement`.
    ([1,1,2,2,3,3,4,4], {1,2}, 5) -> 5, 3, 3, 4, 4
    """

    assert isinstance(col_set, set)

    for key, g in groupby(it, key=operator_in(col_set)):
        if key:
            yield replacement
        else:
            yield from g


def any_in_common(a: Iterable[T], b: Iterable[T]) -> bool:
    """Tests if iterables `a` and `b` have any elements in common.
    See also: `any_in`.
    """

    return any(i == j for i, j in product(a, b))


def any_in(it: Iterable[T], container: Container[T]) -> bool:
    """Tests if any elements of `it` are in `container`.
    For best performance, `container` should be a set.
    See also: `any_in_common`.
    """

    if isinstance(it, set) and isinstance(container, set):
        return not it.isdisjoint(container)

    return any(elm in container for elm in it)


class CountIter:
    """Example:
    c = CountIter()
    with open("asd.txt", "rt", encoding="utf-8") as fr:
            for line in c.count(fr):
                    func(line)  # do something with line
    print("Processed", c.get(), "lines")
    """

    def __init__(self):
        self.i = 0

    def count(self, it):
        for elm in it:
            self.i += 1
            yield elm

    def get(self):
        return self.i


def select_by_indices(it: Iterable[T], indices: Iterator[int]) -> Iterator[T]:
    try:
        sel_ind = next(indices)
        for i, elem in enumerate(it):
            if i == sel_ind:
                yield elem
                sel_ind = next(indices)
    except StopIteration:
        pass


def skip_by_indices(it: Iterator[T], indices: Iterator[int]) -> Iterator[Tuple[int, T]]:
    i = 0
    try:
        sel_ind = next(indices)
        for elem in it:
            if i == sel_ind:
                sel_ind = next(indices)
                continue

            yield i, elem
            i += 1
    except StopIteration:
        pass

    for ii, elem in enumerate(it, i):
        yield ii, elem


def remove_consecutive_dupes(it: Iterable) -> Iterator:
    """Removes consecutive duplicates from iterable `it`.
    (a, a, b, b, a, b) -> a, b, a, b
    """

    return map(itemgetter(0), groupby(it))


def nonstriter(iterable: Iterable) -> Iterator:
    if isinstance(iterable, (str, bytes)):
        raise TypeError("str or bytes")
    else:
        return iter(iterable)


class CachedIterable(Generic[T]):
    # https://stackoverflow.com/a/19504173

    def __init__(self, iterable: Iterable[T]) -> None:
        self.iterable = iterable
        self.done = False
        self.vals: List[T] = []

    def __iter__(self) -> Iterator[T]:
        if self.done:
            return iter(self.vals)
        return chain(self.vals, self._gen_iter())

    def _gen_iter(self) -> Iterator[T]:
        for val in self.iterable:
            self.vals.append(val)
            yield val
        self.done = True


def find_majority_element(it: Iterable[T]) -> Optional[T]:
    """Find most common element of `it` in constant space and linear time (under the assumption,
    that there is indeed one majority element).
    Returns `None` if the iterable is empty.
    """

    counter = 0
    current = None

    for i in it:
        if counter == 0:
            current = i
            counter += 1
        elif i == current:
            counter += 1
        else:
            counter -= 1

    return current


def pairwise_skip(it: Iterable[T], func: Callable[[T], bool]) -> Iterator[Tuple[T, T]]:
    """Yields pairs from iterable `it`, but skips elements where `func(element)` is `True`.
    For example: func = lambda x: x == 2
    (0, 1, 2, 3, 4) -> (0, 1), (3, 4)
    """

    it = iter(it)

    class CustomStopIteration(Exception):
        pass

    def yielduntil():
        try:
            while True:
                elm = next(it)
                if func(elm):
                    return
                yield elm
        except StopIteration:
            raise CustomStopIteration() from None

    try:
        while True:
            yield from pairwise(yielduntil())
    except CustomStopIteration:
        pass


def interleave(*iterables: Iterable) -> Iterator:
    """Interleaves the outputs of multiple iterables into one.
    ((1,3), (2,4)) -> 1,2,3,4
    """

    return chain.from_iterable(zip(*iterables))


def itersort(it: Iterable[T], buffer_size: int) -> Iterator[T]:
    """Partially sorts iterator. Keeps at most `buffer_size` items in memory."""

    heap: List[T] = []
    for item in it:
        if len(heap) == buffer_size:
            yield heapreplace(heap, item)
        else:
            heappush(heap, item)

    while heap:
        yield heappop(heap)


def itershuffle(it: Iterable[T], buffer_size: int) -> Iterator[T]:
    """Partially shuffles `it`. Keeps at most `buffer_size` items in memory."""

    return map(itemgetter(1), itersort(zip(iterrandrange(0, buffer_size), it), buffer_size))


def product_map(func: Callable, *parts: Iterable) -> Iterator:
    for pc in product(*parts):
        yield func(*pc)


# alias: ilen
def iter_len(it: Union[Iterable, Collection]) -> int:
    """Returns the length of iterable `it`.
    If len() is not supported, the iterable will be consumed completely.
    """

    try:
        return len(it)  # type: ignore
    except TypeError:
        return sum(1 for _ in it)



from typing import Callable, Optional


def rename(func_name: str, func_qualname: Optional[str] = None) -> Callable:
    def decorator(func: Callable):
        func.__name__ = func_name
        func.__qualname__ = func_qualname or func_name
        return func

    return decorator


def renameobj(name: str, obj) -> None:
    obj.__class__ = type(name, (obj.__class__,), {})



from itertools import islice
from string import ascii_letters
from typing import Dict, Iterable, Iterator, List, Optional, Sequence, TextIO, Union

from nltk.tokenize import word_tokenize as tokenize
from nltk.tokenize.treebank import TreebankWordDetokenizer

from .file import PathOrTextIO
from .gensim import KeyedVectors

english_word_chars = set(ascii_letters) | {"-", "'"}


def resembles_english_word(word: str) -> bool:
    """Returns True if `word` only consists of English letters and - or '."""

    return all(t in english_word_chars for t in word)


def gensim_indexer(embeddings: KeyedVectors, doc: str, ignore: bool = True) -> Iterator[int]:
    for word in tokenize(doc):
        try:
            yield embeddings.vocab[word.lower()].index
        except KeyError:
            if ignore:
                pass
            else:
                raise


def batch_gensim_indexer(embeddings: KeyedVectors, docs: Iterable[str], ignore: bool = True) -> Iterator[List[int]]:
    for doc in docs:
        yield list(gensim_indexer(embeddings, doc, ignore))


def load_freqs(fname: Union[str, TextIO], normalize: bool = False, limit: Optional[int] = None) -> Dict[str, int]:
    with PathOrTextIO(fname, "rt", encoding="utf-8") as fin:
        freqs = {}

        for line in islice(fin, limit):
            word, count = line.split()
            freqs[word] = int(count)

    if normalize:
        total = sum(freqs.keys())

        for word, count in freqs.items():
            freqs[word] = count / total

    return freqs


def detokenize(tokens: Sequence[str]) -> str:
    """Simply wraps the nltk `TreebankWordDetokenizer` into a convenience function."""

    detokenizer = (
        TreebankWordDetokenizer()
    )  # doesn't use custom `__init__` so should be fast enough to instantiate every time
    s = detokenizer.detokenize(tokens)
    return s



from typing_extensions import Self

from ..exceptions import MalformedFile


class Subtitle:
    def __init__(self, start, end, lines):
        self.start = start
        self.end = end
        self.lines = lines


class Sub:
    """MicroDVD subtitle"""

    sep = "|"

    def __init__(self, path, mode="r", encoding="utf-8-sig"):
        self.fp = open(path, mode, encoding)
        self.current_line = 0

    def __enter__(self) -> Self:
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def _readline(self):
        line = next(self.fp).rstrip()
        self.current_line += 1

        start, end, text = line.split("}", 2)
        try:
            start, end = int(start[1:]), int(end[1:])
        except Exception:
            MalformedFile(f"Error in line {self.current_line}: sub malformed: {line}")

        return Subtitle(start, end, text.split(self.sep))

    def readline(self):
        try:
            return self._readline()
        except StopIteration:
            return None

    def __iter__(self) -> Self:
        return self

    def __next__(self):
        return self._readline()

    def write_subtitles(self, subtitle):
        self.fp.write(f"{{{subtitle.start}}}{{{subtitle.end}}}{self.sep.join(subtitle)}\n")

    def close(self):
        self.fp.close()



from math import pow
from random import sample
from typing import Collection

from .compat.math import prod
from .math import reciprocal


def sample_range(total, num):
    return sample(range(total), min(num, total))


def arithmetic_mean(col: Collection) -> float:
    """Arithmetic mean of `col`."""

    return sum(col) / len(col)


mean = arithmetic_mean


def harmonic_mean(col: Collection) -> float:
    """Harmonic mean of `col`."""

    return len(col) / sum(map(reciprocal, col))


def geometric_mean(col: Collection) -> float:
    """Geometric mean of `col`."""

    return pow(prod(col), reciprocal(len(col)))



import binascii
import gzip
import logging
from hashlib import sha1
from itertools import chain, compress, repeat, zip_longest
from os import fspath
from pathlib import Path
from typing import Any, Dict, Iterable, Iterator, List, Optional, Union

import bencodepy
import requests

from .exceptions import ParseError, assert_choice
from .file import blockfileiter, blockfilesiter
from .filesystem import FileProperties, long_path_support

logger = logging.getLogger(__name__)

BENCODE_BINARY = bencodepy.Bencode(
    encoding=None,
    encoding_fallback=None,
    dict_ordered=False,
    dict_ordered_sort=False,
)
BENCODE_UTF8 = bencodepy.Bencode(encoding="utf-8", encoding_fallback="value", dict_ordered=True, dict_ordered_sort=True)


def read_torrent(path: Path, binary: bool = True) -> dict:
    with path.open("rb") as fr:
        if binary:
            return BENCODE_BINARY.decode(fr.read())
        else:
            return BENCODE_UTF8.decode(fr.read())


def write_torrent(data: dict, path: Path) -> None:
    with path.open("wb") as fw:
        fw.write(BENCODE_BINARY.encode(data))


def iterdecode(items: Iterable[Union[str, bytes]], encoding: str = "latin1") -> Iterator[str]:
    for item in items:
        if isinstance(item, str):
            yield item
        elif isinstance(item, bytes):
            yield item.decode(encoding)
        else:
            raise TypeError(f"items must be strings or bytes, not {type(item)}")


def read_torrent_info_dict(path: Path, normalize_string_fields: bool = False) -> dict:
    if normalize_string_fields:
        info = read_torrent(path, binary=False)["info"]
    else:
        info = read_torrent(path, binary=True)[b"info"]

    if normalize_string_fields:
        if isinstance(info["name"], bytes):
            try:
                info["name"] = info["name.utf-8"]
                del info["name.utf-8"]
                assert isinstance(info["name"], str)
            except KeyError:
                info["name"] = info["name"].decode("latin1")

        try:
            files = info["files"]
        except KeyError:
            pass
        else:
            for fd in files:
                if any(isinstance(p, bytes) for p in fd["path"]):
                    try:
                        fd["path"] = fd["path.utf-8"]
                        del fd["path.utf-8"]
                        assert all(isinstance(p, str) for p in fd["path"])
                    except KeyError:
                        fd["path"] = iterdecode(fd["path"])

    return info


def iter_torrent(path: Path) -> Iterator[FileProperties]:
    """Returns file informations from a torrent file.
    Hash (SHA-1) is obtained according to BEP0047 (if available).
    """

    """ *.utf-8 are hacks used by some clients to add utf-8 encoded strings.
        Now every string should be utf-8 encoded anyway, but in ancient times the encoding was not specified.
    """

    info = read_torrent_info_dict(path, normalize_string_fields=True)

    try:
        files = info["files"]
    except KeyError:
        yield FileProperties(info["name"], info["length"], False, hash=info.get("sha1"))
    else:
        for fd in files:
            yield FileProperties(info["name"] + "/" + "/".join(fd["path"]), fd["length"], False, hash=fd.get("sha1"))


def iter_fastresume(path: Path) -> Iterator[FileProperties]:
    """Files management in libtorrent fastresume files works like this (I think):
    - mapped_files contains a list with new files names, the list is truncated if the tail does not contain renamed files.
    - file_priority contains priorities for each file. Priority 0 means not downloaded.
    """

    try:
        d = read_torrent(path, binary=False)
    except bencodepy.BencodeDecodeError as e:
        logger.warning("%s: %s", path, e)
        raise

    save_path = Path(d["qBt-savePath"])
    save_path_ = Path(d["save_path"])
    assert save_path == save_path_, f"{save_path}, {save_path_}"

    path_torrent = Path(path).with_suffix(".torrent")
    try:
        files = list(iter_torrent(path_torrent))
    except bencodepy.BencodeDecodeError as e:
        logger.warning("%s: %s", path_torrent, e)
        raise

    # file_priority is truncated, remainder should be filled with 1
    file_priority: List[int] = d.get("file_priority", [])
    assert len(file_priority) <= len(files)
    file_priority = list(chain(file_priority, repeat(1, len(files) - len(file_priority))))

    # mapped_files is truncated, remainder should be filled with ""
    mapped_files: List[str] = d.get("mapped_files", [])
    assert len(mapped_files) <= len(files)

    it = compress(zip_longest(files, mapped_files), file_priority)
    for file, mapped_file in it:
        if not mapped_file:  # either None from zip_longest, or "" as placeholder in fastresume file
            assert file.relpath is not None  # for mypy
            file.abspath = long_path_support(fspath(save_path / file.relpath))
        else:
            file.abspath = long_path_support(fspath(save_path / mapped_file))
        file.relpath = None
        yield file


def pieces_field(pieces: Iterable[bytes]) -> bytes:
    return b"".join(sha1(piece).digest() for piece in pieces)  # nosec


def create_torrent_info_dict(
    path: Path, piece_size: int, private: Optional[int] = None, sort: str = "name"
) -> Dict[str, Any]:
    """Create torrent info dict.

    path: Input file or directory
    piece_size: piece size, usually a power of two
    private: set private flag
    sort: sort by file path (name), file size (size) or use abitrary file system order (unordered)
    """

    if private is not None:
        assert_choice("private", private, {0, 1})

    assert_choice("sort", sort, {"name", "size", "unordered"})

    if piece_size < 1:
        raise ValueError("piece_size must be >= 1")

    if path.is_file():
        ret = {
            "name": path.name,
            "length": path.stat().st_size,
            "piece length": piece_size,
            "pieces": pieces_field(blockfileiter(path, chunk_size=piece_size)),
        }

    elif path.is_dir():
        files = list(p for p in path.rglob("*") if p.is_file())

        if sort == "name":
            files = sorted(files, key=lambda path: path.parts)
        elif sort == "size":
            files = sorted(files, key=lambda path: path.stat().st_size)

        assert files, "not implemented yet"

        ret = {
            "files": [
                {
                    "length": file.stat().st_size,
                    "path": file.relative_to(path).parts,
                }
                for file in files
            ],
            "name": path.name,
            "piece length": piece_size,
            "pieces": pieces_field(blockfilesiter(files, chunk_size=piece_size)),
        }

    else:
        raise ValueError("path neither file nor directory")

    if private is not None:
        ret["private"] = private

    return ret


def torrent_info_hash(d: dict) -> str:
    return sha1(BENCODE_BINARY.encode(d)).hexdigest()  # nosec


def get_torrent_hash(path: Path) -> str:
    return torrent_info_hash(read_torrent_info_dict(path))


def create_torrent(
    path: Path, piece_size: int, announce: str = "", private: Optional[bool] = None, sort: str = "name"
) -> bytes:
    info = create_torrent_info_dict(path, piece_size, private, sort)
    torrent = {"announce": announce, "info": info}

    return BENCODE_BINARY.encode(torrent)


def scrape(tracker_url: str, hashes: List[str], timeout: Optional[float] = 120.0) -> Dict[str, dict]:
    if not tracker_url.startswith(("http://", "https://")):
        raise ValueError(f"Only http(s) scrape is supported: <{tracker_url}>")

    base, announce = tracker_url.rsplit("/", 1)
    if "announce" not in announce:
        raise ValueError(f"Scrape not supported for {tracker_url}")

    scrape = announce.replace("announce", "scrape")
    scrape_url = f"{base}/{scrape}"

    hashes = [binascii.a2b_hex(hash) for hash in hashes]

    dec = bencodepy.BencodeDecoder(encoding="utf-8", encoding_fallback="all")

    r = requests.get(scrape_url, params={"info_hash": hashes}, timeout=timeout)
    r.raise_for_status()
    data = r.content
    try:
        tmp = dec.decode(data)
    except bencodepy.BencodeDecodeError:
        tmp = data.rstrip(b"\n")
        try:
            tmp = gzip.decompress(tmp)
        except gzip.BadGzipFile:
            raise ParseError("Failed to parse scrape response", data=data)
        tmp = dec.decode(tmp)

    try:
        files = tmp["files"]
    except KeyError:
        raise ParseError("Missing `files` key in scrape response", data=tmp)

    if len(files) < len(hashes):
        logger.warning("Less hashes returned (%s) than requested (%s)", len(files), len(hashes))

    return {k.hex(): v for k, v in files.items()}


if __name__ == "__main__":
    from argparse import ArgumentParser
    from pprint import pprint

    from genutility.object import compress as _compress

    DEFAULT_ANNOUNCE = "udp://tracker.opentrackr.org:1337/announce"
    DEFAULT_PIECE_LENGTH = 16 * 1024
    DEFAULT_SORT = "name"

    parser = ArgumentParser()
    parser.add_argument("action", choices=("read", "create"))
    parser.add_argument("--in-path", type=Path)
    parser.add_argument("--out-path", type=Path)
    parser.add_argument("--piece-size", type=int, default=DEFAULT_PIECE_LENGTH)
    parser.add_argument("--announce", default=DEFAULT_ANNOUNCE)
    parser.add_argument("--sort", default=DEFAULT_SORT, choices=("name", "size", "unordered"))
    args = parser.parse_args()

    if args.action == "read":
        if args.in_path.suffix == ".torrent":
            path_torrent = args.in_path
            path_fastresume = args.in_path.with_suffix(".fastresume")
        elif args.in_path.suffix == ".fastresume":
            path_torrent = args.in_path.with_suffix(".torrent")
            path_fastresume = args.in_path
        else:
            parser.error("Invalid file extension")

        d = read_torrent(path_torrent, binary=True)
        del d[b"info"][b"pieces"]
        pprint(_compress(d), width=120)
        print()
        d = read_torrent(path_fastresume, binary=False)
        d.pop("pieces", None)
        d.pop("piece_priority", None)
        d.pop("peers", None)
        pprint(_compress(d), width=120)
        print()
        pprint(list(iter_fastresume(path_fastresume)), width=120)

    elif args.action == "create":
        if not args.out_path:
            parser.error("--out-path required for action=create")

        data = create_torrent(args.in_path, args.piece_size, args.announce, sort=args.sort)
        with args.out_path.open("wb") as fw:
            fw.write(data)



from ctypes import FormatError, GetLastError, WinError, byref, sizeof
from ctypes.wintypes import DWORD, USHORT
from errno import EACCES

from cwinsdk.um.fileapi import INVALID_FILE_SIZE, CreateFileW, GetCompressedFileSizeW
from cwinsdk.um.handleapi import INVALID_HANDLE_VALUE
from cwinsdk.um.winbase import GetFileInformationByHandleEx, OpenFileById
from cwinsdk.um.winioctl import FSCTL_SET_COMPRESSION
from cwinsdk.um.winnt import COMPRESSION_FORMAT_DEFAULT, COMPRESSION_FORMAT_NONE, FILE_SHARE_READ, FILE_SHARE_WRITE
from cwinsdk.windows import ERROR_SHARING_VIOLATION  # structs; enums
from cwinsdk.windows import (
    FILE_ID_DESCRIPTOR,
    FILE_ID_INFO,
    FILE_ID_TYPE,
    FILE_INFO_BY_HANDLE_CLASS,
    GENERIC_READ,
    OPEN_EXISTING,
)

from .device import EMPTY_BUFFER, MyDeviceIoControl
from .handle import WindowsHandle, _mode2access

NO_ERROR = 0


class SharingViolation(OSError):
    pass


def GetCompressedFileSize(path: str) -> int:
    HighPart = DWORD()
    LowPart = GetCompressedFileSizeW(path, byref(HighPart))

    if LowPart == INVALID_FILE_SIZE:
        if GetLastError() != NO_ERROR:
            e = WinError()
            e.filename = path
            raise e

    return (HighPart.value << 32) + LowPart


class WindowsFile(WindowsHandle):
    def __init__(self, handle: int) -> None:
        WindowsHandle.__init__(self, handle, doclose=True)

    @classmethod
    def from_path(cls, path: str, mode: str = "r", shared: bool = False) -> "WindowsFile":
        """Create a Windows file objects from `path`.
        If shared is False: allow write access from other processes.
        """

        DesiredAccess = _mode2access[mode]

        if shared:
            ShareMode = FILE_SHARE_READ | FILE_SHARE_WRITE
        else:
            ShareMode = FILE_SHARE_READ

        SecurityAttributes = None
        CreationDisposition = OPEN_EXISTING
        FlagsAndAttributes = 0

        handle = CreateFileW(
            path, DesiredAccess, ShareMode, SecurityAttributes, CreationDisposition, FlagsAndAttributes, None
        )

        if handle == INVALID_HANDLE_VALUE:
            winerror = GetLastError()
            if winerror == ERROR_SHARING_VIOLATION:
                errno = EACCES
                strerror = FormatError(winerror)
                raise SharingViolation(errno, strerror, path, winerror)
            else:
                e = WinError()
                e.filename = path
                raise e

        return cls(handle)

    @classmethod
    def from_fileid(cls, volume, fileid):
        VolumeHint = None  # open volume handle here
        FileId = FILE_ID_DESCRIPTOR(Size=..., Type=FILE_ID_TYPE.ExtendedFileIdType)
        DesiredAccess = GENERIC_READ
        ShareMode = FILE_SHARE_READ | FILE_SHARE_WRITE
        lpSecurityAttributes = None
        FlagsAndAttributes = 0

        handle = OpenFileById(
            VolumeHint, byref(FileId), DesiredAccess, ShareMode, lpSecurityAttributes, FlagsAndAttributes
        )
        return cls(handle)

    def info(self) -> FILE_ID_INFO:
        FileInformation = FILE_ID_INFO()
        GetFileInformationByHandleEx(
            self.handle, FILE_INFO_BY_HANDLE_CLASS.FileIdInfo, byref(FileInformation), sizeof(FileInformation)
        )
        return FileInformation

    def set_compression(self, compressed: bool) -> None:
        if compressed:
            InBuffer = USHORT(COMPRESSION_FORMAT_DEFAULT)
        else:
            InBuffer = USHORT(COMPRESSION_FORMAT_NONE)
        MyDeviceIoControl(self.handle, FSCTL_SET_COMPRESSION, InBuffer, EMPTY_BUFFER())


def is_open_for_write(path: str) -> bool:
    """Tests if file is already open for write
    by trying to open it in exclusive read model.
    """

    try:
        with WindowsFile.from_path(path, mode="r", shared=False):
            return False
    except SharingViolation:
        return True


def set_compression(path: str, compressed: bool) -> None:
    with WindowsFile.from_path(path, mode="r+") as wf:
        wf.set_compression(compressed)


if __name__ == "__main__":
    from argparse import ArgumentParser

    parser = ArgumentParser()
    parser.add_argument("path")
    args = parser.parse_args()

    with WindowsFile.from_path(args.path, mode="r", shared=False) as wf:
        print("Volume serial number:", wf.info().VolumeSerialNumber)
        print("File id:", bytes(wf.info().FileId).hex())



from collections import Counter, defaultdict
from functools import partial
from itertools import islice
from math import factorial, log, pi, sqrt
from operator import gt, lt, mul
from random import sample
from typing import (
    Callable,
    DefaultDict,
    Hashable,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    SupportsFloat,
    Tuple,
    TypeVar,
)

from typing_extensions import Self

from .compat.math import isqrt, prod
from .exceptions import EmptyIterable
from .iter import range_count
from .typing import Computable, Orderable

OrderableT = TypeVar("OrderableT", bound=Orderable)

H = TypeVar("H", bound=Hashable)

inf = float("inf")


def reciprocal(n: Computable) -> Computable:
    return 1.0 / n


def sqr(x: Computable) -> Computable:
    """Returns `x` to the power of 2."""

    return x * x


def dot(a: Iterable[Computable], b: Iterable[Computable]) -> Computable:
    """Returns the dot product (inner product) of vectors `a` and `b`."""

    return sum(map(mul, a, b))


# was: euclidic_norm
def euclidean_norm(x: Iterable[Computable]) -> Computable:
    """Returns the Euclidean norm of vector `x`."""

    return sqrt(sum(map(sqr, x)))


def shannon_entropy(ps: Iterable[float], base: int = 2) -> float:
    """Calculates the Shannon entropy for probabilities `ps` with `base` in pure Python."""

    return -sum(p * log(p, base) for p in ps)


def cosine_similarity(a: Iterable[Computable], b: Iterable[Computable]) -> Computable:
    """Calculate the cosine similarity (normalized dot product) of vectors `a` and `b`."""

    return dot(a, b) / euclidean_norm(a) / euclidean_norm(b)


def argmin(seq: Sequence[OrderableT], s: int = 0, e: Optional[int] = None) -> OrderableT:
    if not seq:
        raise EmptyIterable()

    arg = s
    e = e or len(seq)
    for i in range(s + 1, e):
        if seq[i] < seq[arg]:
            arg = i
    return arg


def argmax(seq: Sequence[OrderableT], s: int = 0, e: Optional[int] = None) -> int:
    if not seq:
        raise EmptyIterable()

    arg = s
    e = e or len(seq)
    for i in range(s + 1, e):
        if seq[i] > seq[arg]:
            arg = i
    return arg


def argmax_pair(iterable: Iterable[OrderableT]) -> Tuple[int, OrderableT]:
    it = iter(iterable)
    arg = 0
    try:
        max = next(it)
    except StopIteration:
        raise EmptyIterable("empty iterable") from None

    for i, elm in enumerate(it, 1):
        if elm > max:
            arg = i
            max = elm

    return arg, max


from itertools import count
from operator import itemgetter


def argmax_v2(iterable: Iterable[OrderableT]) -> int:
    """nicer, but almost 2 times slower than above"""
    return max(zip(count(), iterable), key=itemgetter(1))[0]


def minmax(a: OrderableT, b: OrderableT) -> Tuple[OrderableT, OrderableT]:
    """`default` argument cannot be used, because the C level function doesn't have a default value
    but an overload.
    """

    if a <= b:
        return a, b
    else:
        return b, a


def _argfind_cmp(
    it: Iterable[OrderableT],
    target: OrderableT,
    op1: Callable[[OrderableT, OrderableT], bool],
    op2: Callable[[OrderableT, OrderableT], bool],
    pos: int = 0,
) -> Tuple[int, Optional[OrderableT]]:
    """Find the largest element in `it` less than or equal to `target` and return the index and value.
    Return (-1, None) if no such element can be found.
    """

    idx = -1
    val: Optional[OrderableT] = None

    it = islice(it, pos, None)
    try:
        v = next(it)
    except StopIteration:
        return idx, val

    if v == target:
        return pos, v
    elif op1(v, target):
        idx = pos
        val = v

    for i, v in enumerate(it, pos + 1):
        if v == target:
            return i, v
        elif op1(v, target) and (val is None or op2(v, val)):
            idx = i
            val = v

    return idx, val


def argfind_lte(it, target, pos=0):
    return _argfind_cmp(it, target, lt, gt, pos)


def argfind_gte(it, target, pos=0):
    return _argfind_cmp(it, target, gt, lt, pos)


def degree_to_rad(angle: float) -> float:
    return 2.0 * pi * angle / 360.0


def limit(x, left=0, right=1):
    if x >= left:
        if x <= right:
            return x
        else:
            return right
    else:
        return left


def relative_luminance(r: float, g: float, b: float) -> float:
    """Converts linear RGB components to relative luminance.
    See: https://en.wikipedia.org/wiki/Relative_luminance
    """

    return 0.2126 * r + 0.7152 * g + 0.0722 * b


class _PosInfInt:
    """A fancy version of `sys.maxsize`"""

    def __lt__(self, rhs) -> bool:
        return False

    def __gt__(self, rhs) -> bool:
        return True

    def __eq__(self, rhs) -> bool:
        if isinstance(rhs, _PosInfInt):
            raise ArithmeticError("Cannot compare inf with inf")

        return False

    def __add__(self, rhs) -> Self:
        return self

    def __sub__(self, rhs) -> Self:
        if isinstance(rhs, _PosInfInt):
            raise ArithmeticError("Cannot subtract inf from inf")

        return self

    def __str__(self) -> str:
        return "PosInfInt"

    def __repr__(self) -> str:
        return "PosInfInt"


PosInfInt = _PosInfInt()


def multinomial_coefficient(n: int, ks: Iterable[int]) -> int:
    return factorial(n) / prod(map(factorial, ks))


def num_unique_permutations(word: str) -> int:
    """Number of unique permutations of `word`."""

    return multinomial_coefficient(len(word), Counter(word).values())


def discrete_distribution(it: Iterable[H]) -> Tuple[DefaultDict[H, int], int]:
    """Creates a discrete distribution of the elements of `it`."""

    d: DefaultDict[H, int] = defaultdict(int)
    s = 0
    for i in it:
        s += 1
        d[i] += 1
    return d, s


def fibonaccigen(f0: int = 0, f1: int = 1) -> Iterator[int]:
    """A000045 [0, 1, 1, 2, 3, 5, 8, 13, 21, 34, 55, ...]
    arguments are the first two element of the Fibonacci series

    To get a specific Fibonacci number use `fibonacci(n)`.
    """

    yield f0
    while True:
        yield f1
        f1, f0 = f1 + f0, f1


def floordiv2(n: int) -> Iterator[int]:
    while n > 0:
        yield n
        n = n // 2


def isintlog(num: int, base: int) -> bool:
    if num <= 0 or base <= 0:
        raise ValueError("num and base must be larger than 0")

    if num == 1:
        return False

    while num > 1:
        r = num % base
        if r != 0:
            return False
        num //= base

    return True


def _fibonacci(n: int) -> Tuple[int, int]:
    # returns (F(n), F(n+1)).
    # inspired by https://www.nayuki.io/page/fast-fibonacci-algorithms

    a, b = 0, 1
    for i in reversed(tuple(floordiv2(n))):
        c = a * (b * 2 - a)
        d = a * a + b * b
        if i % 2 == 0:
            a = c
            b = d
        else:
            a = d
            b = c + d
    return a, b


# from .numba import opjit
# @opjit() # doesn't support arbitrary large ints
def _fibonacci_v2(n: int) -> Tuple[int, int]:
    # see: https://stackoverflow.com/a/1526036
    # 'Structure and Interpretation of Computer Programs' (SICP) ex 1.19
    # http://community.schemewiki.org/?sicp-ex-1.19

    a, b, p, q = 1, 0, 0, 1
    while n > 0:
        if n % 2 == 0:  # even
            oldp = p
            p = p * p + q * q
            q = 2 * oldp * q + q * q
            n //= 2
        else:
            olda = a
            a = b * q + a * q + a * p
            b = b * p + olda * q
            n -= 1
    return a, b


def fibonacci(n: int) -> int:
    """implementation of "fast doubling". can be derived from this identity:
    [[1, 1], [1, 0]]^n = [[F_n+1, F_n], [F_n, F_n-1]]
    """

    if n == 0:
        return 0
    elif n < 0:
        raise ValueError("Negative arguments not implemented")

    return _fibonacci(n - 1)[1]


def byte2size(byte: SupportsFloat, exp: int = 0, base: int = 1024) -> Tuple[float, str]:
    """Converts integer number to float and unit string"""

    byte = float(byte)
    byte_units = ("Byte", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB", "YiB")

    for unit in islice(byte_units, exp, None):  # noqa: B007
        if byte >= base:
            byte /= base
        else:
            break

    return byte, unit


def byte2size_str(byte: SupportsFloat, roundval: int = 3) -> str:
    """Converts integer number to human readable size string."""

    byte, value = byte2size(byte)
    if value == "Byte":
        roundval = 0
    return "{:.{}f} {}".format(byte, roundval, value)


def primes(stop: Optional[int] = None) -> Iterable[int]:
    """Naive algorithm to yield all primes which are smaller than `stop`.
    If `stop` is None it will never stop generating primes.
    For better algorithms see here: https://stackoverflow.com/questions/2211990/how-to-implement-an-efficient-infinite-generator-of-prime-numbers-in-python/10733621#10733621
    """

    found_primes: List[int] = []
    first_prime = 2

    for i in range_count(first_prime, stop):
        is_prime = True
        for j in found_primes:
            if j > isqrt(i):
                break
            if i % j == 0:
                is_prime = False
                break

        if is_prime:
            found_primes.append(i)
            yield i


def additaet(n: int) -> int:
    """sums all numbers from 0 to n using Gaussian formula.
    basically same performance
            (n*n + n) >> 1
    too slow, floats
            int(((n/2.0) + 0.5) * n)
            int((n + 1) / 2.0 * n)
    too slow, worse complexity
            sum(range(n+1))
    """

    return ((n + 1) * n) >> 1


addity = additaet


def digitsum(n: int) -> int:
    """Use for numbers with more than 6 digits."""

    return sum(map(int, str(n)))


def digitsum_small(n: int) -> int:
    """Use for numbers with less than 6 digits."""

    s = 0
    while n:
        s += n % 10
        n //= 10
    return s


def digitsum_base(n: int, base: int = 10) -> int:
    """Use with string input if base is not 10."""

    return sum(map(partial(int, base=base), str(n)))


def euclidean_distance(a: Computable, b: Computable) -> Computable:
    """Euclidean distance between `a` and `b`."""

    return abs(b - a)


def closest(
    numbers: Iterable[Computable],
    number: Computable,
    distance_func: Callable[[Computable, Computable], Computable] = euclidean_distance,
) -> Computable:
    """For a list of `numbers`, return the closest number to `number`."""

    return min(numbers, key=partial(distance_func, number))


def absolute_difference(a: Computable, b: Computable) -> Computable:
    return abs(b - a)


def sortedsample(n: int, k: int) -> List[int]:
    """Sorted random sample without replacements."""

    return sorted(sample(range(n), k))


def int2base(x: int, base: int) -> str:
    """Converts number `x` to its string representation in `base`."""

    import string

    assert base > 1
    digs = string.digits + string.ascii_letters

    if x < 0:
        sign = -1
    elif x == 0:
        return digs[0]
    else:
        sign = 1

    x *= sign
    digits = []

    while x:
        digits.append(digs[x % base])
        x = x // base

    if sign < 0:
        digits.append("-")

    return "".join(reversed(digits))


number_metric = absolute_difference

if __name__ == "__main__":
    import timeit

    print(min(timeit.repeat(stmt="digitsum(111111)", setup="from __main__ import digitsum")))
    print(min(timeit.repeat(stmt="digitsum_base(111111)", setup="from __main__ import digitsum_base")))
    # print(min(timeit.repeat('_fibonacci_v2(999999999999999999999999999999999)', setup="from __main__ import _fibonacci_v2")))



import fcntl
import os
import signal
from typing import IO

from .os_shared import _usagetuple


def _lock(fp: IO, exclusive: bool = True, block: bool = False) -> None:
    if exclusive:
        operation = fcntl.LOCK_EX
    else:
        operation = fcntl.LOCK_SH

    if not block:
        operation |= fcntl.LOCK_NB

    fcntl.flock(fp, operation)


def _unlock(fp: IO) -> None:
    fcntl.flock(fp, fcntl.LOCK_UN)


def _disk_usage_posix(path: str) -> _usagetuple:
    st = os.statvfs(path)

    total = st.f_blocks * st.f_frsize
    free = st.f_bavail * st.f_frsize

    return _usagetuple(total, total - free, free)


def _interrupt_posix() -> None:
    os.kill(os.getpid(), signal.SIGINT)


def _filemanager_cmd_posix(path: str) -> str:
    return f'nautilus "{path}"'  # gnome only. xdg-open for the rest?


def _get_appdata_dir(roaming: bool = False) -> str:
    return os.getenv("XDG_DATA_HOME", os.path.expanduser("~/.local/share"))
